# relative
from ..object_delete.object_delete_message import ObjectDeleteMessage  # noqa: F401
from .upload_service_messages import AbortDataUploadMessage  # noqa: F401
from .upload_service_messages import UploadDataCompleteMessage  # noqa: F401
from .upload_service_messages import UploadDataMessage  # noqa: F401



# stdlib
from typing import Optional

# third party
from nacl.signing import VerifyKey

# relative
from ... import UID
from ......logger import info
from ....abstract.node import AbstractNode
from .simple_messages import NodeRunnableMessageWithReply


class DoesObjectExistMessage(NodeRunnableMessageWithReply):

    __attr_allowlist__ = ["obj_id"]

    def __init__(self, obj_id: UID) -> None:
        self.obj_id = obj_id

    def run(self, node: AbstractNode, verify_key: Optional[VerifyKey] = None) -> bool:
        try:
            return bool(node.store.get_or_none(self.obj_id, proxy_only=True))  # type: ignore
        except Exception as e:
            info("Exception in DoesObjectExistMessage:" + str(e))
            return False



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/torch/module.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.lib.python import dict_pb2 as proto_dot_lib_dot_python_dot_dict__pb2
from syft.proto.lib.python import list_pb2 as proto_dot_lib_dot_python_dot_list__pb2
from syft.proto.lib.python.collections import (
    ordered_dict_pb2 as proto_dot_lib_dot_python_dot_collections_dot_ordered__dict__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1cproto/lib/torch/module.proto\x12\x0esyft.lib.torch\x1a/proto/lib/python/collections/ordered_dict.proto\x1a\x1bproto/lib/python/dict.proto\x1a\x1bproto/lib/python/list.proto"\xef\x03\n\x06Module\x12\x13\n\x0bmodule_type\x18\x01 \x01(\t\x12\x13\n\x0bmodule_name\x18\x02 \x01(\t\x12\x18\n\x0bmodule_repr\x18\x03 \x01(\tH\x00\x88\x01\x01\x12\x15\n\rpython_module\x18\x04 \x01(\t\x12(\n\x08\x63hildren\x18\x05 \x03(\x0b\x32\x16.syft.lib.torch.Module\x12/\n\x0bmodule_args\x18\x06 \x01(\x0b\x32\x15.syft.lib.python.ListH\x01\x88\x01\x01\x12\x31\n\rmodule_kwargs\x18\x07 \x01(\x0b\x32\x15.syft.lib.python.DictH\x02\x88\x01\x01\x12<\n\nstate_dict\x18\x08 \x01(\x0b\x32(.syft.lib.python.collections.OrderedDict\x12<\n\nparameters\x18\t \x01(\x0b\x32(.syft.lib.python.collections.OrderedDict\x12@\n\t_uid2attr\x18\x0b \x01(\x0b\x32(.syft.lib.python.collections.OrderedDictH\x03\x88\x01\x01\x42\x0e\n\x0c_module_reprB\x0e\n\x0c_module_argsB\x10\n\x0e_module_kwargsB\x0c\n\nX_uid2attrb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.lib.torch.module_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _MODULE._serialized_start = 156
    _MODULE._serialized_end = 651
# @@protoc_insertion_point(module_scope)



# stdlib
from copy import deepcopy
from functools import reduce
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union
from uuid import UUID

# third party
import numpy as np

# relative
from ....common.uid import UID
from ....store.storeable_object import StorableObject
from ....tensor.autodp.phi_tensor import PhiTensor
from ....tensor.smpc import context
from ....tensor.smpc import utils
from ....tensor.smpc.share_tensor import ShareTensor
from ....tensor.smpc.utils import count_wraps
from ...abstract.node import AbstractNode
from .beaver_action import BeaverAction
from .greenlets_switch import beaver_retrieve_object
from .greenlets_switch import crypto_store_retrieve_object


def get_id_at_location_from_op(seed: int, operation_str: str) -> UID:
    generator = np.random.default_rng(seed)
    return UID(UUID(bytes=generator.bytes(16)))


def private_mul(x: ShareTensor, y: ShareTensor, op_str: str) -> ShareTensor:
    """Performs SMPC Private Multiplication using Beaver Triples.

    Args:
        x (ShareTensor): input share tensor
        y (ShareTensor): input share tensor

    Returns:
        res (ShareTensor): Result of the operation.
    """
    seed_id_locations = context.SMPC_CONTEXT.get("seed_id_locations", None)
    if seed_id_locations is None:
        raise ValueError(
            f"Input seed : {seed_id_locations} for private multiplication should not None"
        )
    generator = np.random.default_rng(seed_id_locations)
    _ = UID(
        UUID(bytes=generator.bytes(16))
    )  # Ignore first one as it is used for result.
    eps_id = UID(UUID(bytes=generator.bytes(16)))
    delta_id = UID(UUID(bytes=generator.bytes(16)))
    ring_size = utils.get_ring_size(x.ring_size, y.ring_size)
    a_shape = tuple(x.shape)
    b_shape = tuple(y.shape)

    # TODO: Make beaver retrieval context less, this might fail when there are several
    # parallel multiplications.
    a_share, b_share, c_share = crypto_store_retrieve_object(
        f"beaver_{op_str}",
        a_shape=a_shape,
        b_shape=b_shape,
        ring_size=ring_size,
        remove=True,
    )

    node = context.SMPC_CONTEXT.get("node", None)
    # SMPC Multiplication

    # Phase 1: Communication Phase (Beaver dispatch)
    spdz_mask(x, y, eps_id, delta_id, a_share, b_share, node)

    # Phase 2: Share Reconstruction Phase:
    res = spdz_multiply(x, y, op_str, eps_id, delta_id, a_share, b_share, c_share, node)

    return res


def spdz_mask(
    x: ShareTensor,
    y: ShareTensor,
    eps_id: UID,
    delta_id: UID,
    a_share: ShareTensor,
    b_share: ShareTensor,
    node: Optional[AbstractNode] = None,
) -> None:

    if node is None:
        raise ValueError("Node context should be passed to spdz mask")

    clients = ShareTensor.login_clients(parties_info=x.parties_info)

    eps = x - a_share  # beaver intermediate values.
    delta = y - b_share

    client_id_map = {client.id: client for client in clients}
    curr_client = client_id_map[node.id]  # type: ignore
    beaver_action = BeaverAction(
        values=[eps, delta],
        locations=[eps_id, delta_id],
        address=curr_client.address,
    )
    beaver_action.execute_action(node, None)

    for _, client in enumerate(clients):
        if client != curr_client:
            beaver_action.address = client.address
            client.send_immediate_msg_without_reply(msg=beaver_action)


def spdz_multiply(
    x: ShareTensor,
    y: ShareTensor,
    op_str: str,
    eps_id: UID,
    delta_id: UID,
    a_share: ShareTensor,
    b_share: ShareTensor,
    c_share: ShareTensor,
    node: Optional[Any] = None,
) -> ShareTensor:

    nr_parties = x.nr_parties
    ring_size = x.ring_size

    eps = beaver_retrieve_object(node, eps_id, nr_parties)  # type: ignore
    delta = beaver_retrieve_object(node, delta_id, nr_parties)  # type: ignore

    eps: ShareTensor = sum(eps.data)  # type: ignore
    delta: ShareTensor = sum(delta.data)  # type: ignore

    op = ShareTensor.get_op(ring_size, op_str)
    add_op = ShareTensor.get_op(ring_size, "add")

    eps_b = op(eps.child, b_share.child)  # type: ignore
    delta_a = op(a_share.child, delta.child)  # type: ignore

    tensor = reduce(add_op, [c_share.child, eps_b, delta_a])
    if x.rank == 0:
        eps_delta = op(eps.child, delta.child)  # type: ignore
        tensor = add_op(tensor, eps_delta)

    share = x.copy_tensor()
    share.child = tensor

    return share


def public_divide(x: ShareTensor, y: Union[int, np.integer]) -> ShareTensor:
    """Performs SMPC Public Division.

    Args:
        x (ShareTensor): input share tensor
        y (Union[int, np.integer]): input share tensor

    Returns:
        res (ShareTensor): Result of the operation.
    """
    seed_id_locations = context.FPT_CONTEXT.get("seed_id_locations", None)
    if seed_id_locations is None:
        raise ValueError(
            f"Input seed : {seed_id_locations} for public  division should not None"
        )
    generator = np.random.default_rng(seed_id_locations)
    # SKIPPING first 3 location as they are used for multiplication.
    for _ in range(3):
        _ = UID(
            UUID(bytes=generator.bytes(16))
        )  # Ignore first one as it is used for result.
    z_id = UID(UUID(bytes=generator.bytes(16)))  # id for result of x+r

    ring_size = x.ring_size
    shape = tuple(x.shape)

    # TODO: Make  retrieval context less, this might fail when there are several
    # parallel division.
    r_sh, theta_r_sh = crypto_store_retrieve_object(
        "beaver_wraps",
        shape=shape,
        ring_size=ring_size,
        remove=True,
    )

    node = context.SMPC_CONTEXT.get("node", None)
    # SMPC Public Division

    # Phase 1: Communication Phase (Beaver dispatch)
    divide_mask(x, r_sh, z_id, node)

    # Phase 2: Share Corretion Phase:
    theta_x = divide_wrap_correction(x, y, z_id, r_sh, theta_r_sh, node)

    x.child = np.trunc(x.child / y).astype(x.child.dtype)
    # To avoid out-of-bounds when y is small
    res = x - theta_x * 4 * ((x.ring_size // 4) // y)

    return res


def divide_mask(
    x: ShareTensor,
    r: ShareTensor,
    z_id: UID,
    node: Optional[AbstractNode] = None,
) -> None:

    if node is None:
        raise ValueError("Node context should be passed to spdz mask")

    clients = ShareTensor.login_clients(parties_info=x.parties_info)

    z = x + r

    client_id_map = {client.id: client for client in clients}
    curr_client = client_id_map[node.id]  # type: ignore
    beaver_action = BeaverAction(
        values=[z],
        locations=[z_id],
        address=curr_client.address,
    )
    beaver_action.execute_action(node, None)

    for _, client in enumerate(clients):
        if client != curr_client:
            beaver_action.address = client.address
            client.send_immediate_msg_without_reply(msg=beaver_action)


def divide_wrap_correction(
    x: ShareTensor,
    y: Union[int, np.integer],
    z_id: UID,
    r_share: ShareTensor,
    theta_r_sh: ShareTensor,
    node: Optional[AbstractNode] = None,
) -> ShareTensor:
    """Privately computes the number of wraparounds for a set a shares.
    Adapted From Crypten

    To do so, we note that:
        [theta_x] = theta_z + [beta_xr] - [theta_r] - [eta_xr]
    Where:
        [theta_x] is the wraps for a variable x
        [beta_xr] is the differential wraps for variables x and r
        [eta_xr]  is the plaintext wraps for variables x and r
    Note: Since [eta_xr] = 0 with probability 1 - |x| / Q for modulus Q, we
    can make the assumption that [eta_xr] = 0 with high probability.

    Args:
        x (ShareTensor): shares for which we want to compute the number of wraparounds
        y (Union[int, np.integer]): the number/tensor by which we divide
        r_share (ShareTensor): share for a random variable "r"
        theta_r_sh (ShareTensor): share for the number of wraparounds for "r"
        z_id (UID): UID which contains the z(z=x+r_share) from all parties
        node (Optional[Any]) : current node context.

    Returns:
        ShareTensor representing the number of wraparounds
    """
    nr_parties = x.nr_parties
    theta_x = x.copy_tensor()

    z_shares = beaver_retrieve_object(node, z_id, nr_parties).data  # type: ignore

    beta_xr = count_wraps([x.child, r_share.child])

    theta_x.child = beta_xr - theta_r_sh.child
    theta_z = count_wraps([share.child for share in z_shares])

    # TODO: We neglect calculation of Etaxr as it involves a comparision which bottlnecks
    # the computation, we assume Etaxr =0 , the probability of an error in the division
    # is low when we make this assumption.
    theta_x = theta_x + theta_z

    return theta_x


def _decomposition(x: ShareTensor, ring_size: int, bitwise: bool) -> None:
    seed_id_locations = context.SMPC_CONTEXT.get("seed_id_locations", None)
    node = context.SMPC_CONTEXT.get("node", None)
    read_permissions = context.SMPC_CONTEXT.get("read_permissions", None)
    if seed_id_locations is None or node is None or read_permissions is None:
        raise ValueError(
            f"Input values seed,node,read_permissions : {seed_id_locations,node,read_permissions}"
            + " for decomposition should not None"
        )

    local_decomposition(
        x, ring_size, bitwise, seed_id_locations, node, read_permissions
    )


def local_decomposition(
    x: ShareTensor,
    ring_size: int,
    bitwise: bool,
    seed_id_locations: int,
    node: Any,
    read_permissions: Dict[Any, Any],
) -> None:
    """Performs local decomposition to generate shares of shares.

    Args:
        x (ShareTensor) : input ShareTensor.
        ring_size (str) : Ring size to generate decomposed shares in.
        bitwise (bool): Perform bit level decomposition on bits if set.

    Returns:
        List[List[ShareTensor]]: Decomposed shares in the given ring size.
    """
    generator = np.random.default_rng(seed_id_locations)
    # Skip the first ID ,as it is used for None return type in run class method.
    _ = UID(UUID(bytes=generator.bytes(16)))

    tensor_values = context.tensor_values if context.tensor_values else None
    context.tensor_values = None

    rank = x.rank
    nr_parties = x.nr_parties
    numpy_type = utils.RING_SIZE_TO_TYPE[ring_size]
    shape = x.shape
    zero = np.zeros(shape, numpy_type)

    input_shares = []

    if bitwise:
        ring_bits = utils.get_nr_bits(x.ring_size)  # for bit-wise decomposition
        input_shares = [x.bit_extraction(idx) for idx in range(ring_bits)]
    else:
        input_shares.append(x)

    for share in input_shares:

        for i in range(nr_parties):
            id_at_location = UID(UUID(bytes=generator.bytes(16)))
            sh = x.copy_tensor()
            sh.ring_size = ring_size
            if rank != i:
                sh.child = deepcopy(zero)
            else:
                sh.child = deepcopy(share.child.astype(numpy_type))

            if tensor_values is not None:
                if isinstance(tensor_values.child, PhiTensor):
                    # Remove it when we move PC to share tensor.
                    if ring_size != 2:  # type: ignore
                        sh.child = tensor_values.child.child.encode(sh.child)
                    tensor_values.child.child.child = sh
                else:
                    tensor_values.child = sh
                data = tensor_values
            else:
                data = sh

            store_obj = StorableObject(
                id=id_at_location,
                data=data,
                read_permissions=read_permissions,
            )

            node.store[id_at_location] = store_obj


# Function mapped with seed in RunClassMethodSMPC
ACTION_FUNCTIONS = ["__add__", "__sub__", "__mul__", "bit_decomposition"]

# #############################################################
# ### NOTE: DO NOT DELETE COMMENTED CODE,TO BE USED LATER #####
# #############################################################


# ##########ADD,SUB OPS################
# def smpc_basic_op(
#     op_str: str,
#     nr_parties: int,
#     self_id: UID,
#     other_id: UID,
#     seed_id_locations: int,
#     node: Any,
# ) -> List[SMPCActionMessage]:
#     # relative
#     from ..... import Tensor

#     """Generator for SMPC public/private operations add/sub"""

#     generator = np.random.default_rng(seed_id_locations)
#     result_id = UID(UUID(bytes=generator.bytes(16)))
#     other = node.store[other_id].data

#     actions = []
#     if isinstance(other, (ShareTensor, Tensor)):
#         # All parties should add the other share if empty list
#         actions.append(
#             SMPCActionMessage(
#                 f"mpc_{op_str}",
#                 self_id=self_id,
#                 args_id=[other_id],
#                 kwargs_id={},
#                 ranks_to_run_action=list(range(nr_parties)),
#                 result_id=result_id,
#                 address=node.address,
#             )
#         )
#     else:
#         actions.append(
#             SMPCActionMessage(
#                 "mpc_noop",
#                 self_id=self_id,
#                 args_id=[],
#                 kwargs_id={},
#                 ranks_to_run_action=list(range(1, nr_parties)),
#                 result_id=result_id,
#                 address=node.address,
#             )
#         )

#         # Only rank 0 (the first party) would do the add/sub for the public value
#         actions.append(
#             SMPCActionMessage(
#                 f"mpc_{op_str}",
#                 self_id=self_id,
#                 args_id=[other_id],
#                 kwargs_id={},
#                 ranks_to_run_action=[0],
#                 result_id=result_id,
#                 address=node.address,
#             )
#         )

#     return actions


# def get_action_generator_from_op(
#     operation_str: str, nr_parties: int
# ) -> Callable[[UID, UID, int, Any], Any]:
#     """ "
#     Get the generator for the operation provided by the argument
#     Arguments:
#         operation_str (str): the name of the operation

#     """
#     return functools.partial(MAP_FUNC_TO_ACTION[operation_str], nr_parties)


# # Given an SMPC Action map it to an action constructor
# MAP_FUNC_TO_ACTION: Dict[
#     str, Callable[[int, UID, UID, int, Any], List[SMPCActionMessage]]
# ] = {
#     "__add__": functools.partial(smpc_basic_op, "add"),
#     "__sub__": functools.partial(smpc_basic_op, "sub"),
#     "__mul__": smpc_mul,  # type: ignore
#     "bit_decomposition": bit_decomposition,  # type: ignore
#     # "__gt__": smpc_gt,  # type: ignore TODO: this should be added back when we have only one action
# }


# # Map given an action map it to a function that should be run on the shares"
# _MAP_ACTION_TO_FUNCTION: Dict[str, Callable[..., Any]] = {
#     "mpc_add": functools.partial(apply_function, op_str="add"),
#     "mpc_sub": functools.partial(apply_function, op_str="sub"),
#     "mpc_mul": functools.partial(apply_function, op_str="mul"),
#     "spdz_mask": spdz_mask,
#     "spdz_multiply": spdz_multiply,
#     "local_decomposition": local_decomposition,
#     "mpc_noop": deepcopy,
# }


# ###################PRIVATE MUL#############3
# def smpc_mul(
#     nr_parties: int,
#     self_id: UID,
#     other_id: UID,
#     seed_id_locations: Optional[int] = None,
#     node: Optional[Any] = None,
# ) -> SMPCActionSeqBatchMessage:
#     """Generator for the smpc_mul with a public value"""
#     # relative
#     from ..... import Tensor

#     if seed_id_locations is None or node is None:
#         raise ValueError(
#             f"The values seed_id_locations{seed_id_locations}, Node:{node} should not be None"
#         )
#     generator = np.random.default_rng(seed_id_locations)
#     result_id = UID(UUID(bytes=generator.bytes(16)))
#     _self = node.store[self_id].data
#     other = node.store[other_id].data

#     actions = []
#     if isinstance(other, (ShareTensor, Tensor)):

#         if isinstance(other, ShareTensor):
#             ring_size = other.ring_size
#         else:
#             ring_size = other.child.child.ring_size

#         mask_result = UID(UUID(bytes=generator.bytes(16)))
#         eps_id = UID(UUID(bytes=generator.bytes(16)))
#         delta_id = UID(UUID(bytes=generator.bytes(16)))
#         a_shape = tuple(_self.shape)
#         b_shape = tuple(other.shape)

#         a_share, b_share, c_share = crypto_store_retrieve_object(
#             "beaver_mul",
#             a_shape=a_shape,
#             b_shape=b_shape,
#             ring_size=ring_size,
#             remove=True,
#         )

#         actions.append(
#             SMPCActionMessage(
#                 "spdz_mask",
#                 self_id=self_id,
#                 args_id=[other_id],
#                 kwargs_id={},
#                 kwargs={
#                     "eps_id": eps_id,
#                     "delta_id": delta_id,
#                     "a_share": a_share,
#                     "b_share": b_share,
#                     "c_share": c_share,
#                 },
#                 ranks_to_run_action=list(range(nr_parties)),
#                 result_id=mask_result,
#                 address=node.address,
#             )
#         )

#         actions.append(
#             SMPCActionMessage(
#                 "spdz_multiply",
#                 self_id=self_id,
#                 args_id=[other_id],
#                 kwargs_id={},
#                 kwargs={
#                     "eps_id": eps_id,
#                     "delta_id": delta_id,
#                     "a_share": a_share,
#                     "b_share": b_share,
#                     "c_share": c_share,
#                 },
#                 ranks_to_run_action=list(range(nr_parties)),
#                 result_id=result_id,
#                 address=node.address,
#             )
#         )

#     else:
#         # All ranks should multiply by that public value
#         actions.append(
#             SMPCActionMessage(
#                 "mpc_mul",
#                 self_id=self_id,
#                 args_id=[other_id],
#                 kwargs_id={},
#                 ranks_to_run_action=list(range(nr_parties)),
#                 result_id=result_id,
#                 address=node.address,
#             )
#         )
#     batch = SMPCActionSeqBatchMessage(smpc_actions=actions, address=node.address)
#     return batch


# ###################### BIT DECOMPOSITION ####################
# def bit_decomposition(
#     nr_parties: int,
#     self_id: UID,
#     ring_size: UID,
#     bitwise: UID,
#     seed_id_locations: int,
#     node: Any,
# ) -> List[SMPCActionMessage]:
#     generator = np.random.default_rng(seed_id_locations)
#     result_id = UID(UUID(bytes=generator.bytes(16)))

#     actions = []

#     actions.append(
#         SMPCActionMessage(
#             "local_decomposition",
#             self_id=self_id,
#             args_id=[ring_size, bitwise],
#             kwargs_id={},
#             kwargs={"seed_id_locations": str(seed_id_locations)},
#             ranks_to_run_action=list(range(nr_parties)),
#             result_id=result_id,
#             address=node.address,
#         )
#     )

#     return actions


# ###########################COMPARISON#####################
#
# def smpc_gt(
#     nr_parties: int,
#     self_id: UID,
#     other_id: UID,
#     seed_id_locations: int,
#     node: Any,
#     client: Any,
# ) -> List[SMPCActionMessage]:
#     """Generator for the smpc_mul with a public value"""
#     generator = np.random.default_rng(seed_id_locations)

#     result_id = UID(UUID(bytes=generator.bytes(16)))
#     sub_result = UID(UUID(bytes=generator.bytes(16)))

#     x = node.store.get(self_id).data  # noqa
#     y = node.store.get(other_id).data

#     if not isinstance(y, ShareTensor):
#         raise ValueError("Only private compare works at the moment")

#     actions = []
#     actions.append(
#         SMPCActionMessage(
#             "mpc_sub",
#             self_id=self_id,
#             args_id=[other_id],
#             kwargs_id={},
#             ranks_to_run_action=list(range(nr_parties)),
#             result_id=sub_result,
#             address=client.address,
#         )
#     )

#     actions.append(
#         SMPCActionMessage(
#             "bit_decomposition",
#             self_id=sub_result,
#             args_id=[],
#             # TODO: This value needs to be changed to something else and probably used
#             # directly the przs_generator from ShareTensor - check bit_decomposition function
#             kwargs_id={},
#             ranks_to_run_action=list(range(nr_parties)),
#             result_id=result_id,
#             address=client.address,
#         )
#     )
#     return actions



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import KeysView
from typing import Tuple
from typing import Union

# third party
from nacl.encoding import HexEncoder
from nacl.signing import VerifyKey
from pydantic import BaseModel

# relative
from ......core.node.common.exceptions import InvalidNodeCredentials


class NodeCredentials(BaseModel):
    node_uid: str
    node_name: str
    node_type: str
    verify_key: str

    # allows splatting with **payload
    def keys(self) -> KeysView[str]:
        return self.__dict__.keys()

    # allows splatting with **payload
    def __getitem__(self, key: str) -> Any:
        return self.__dict__.__getitem__(key)

    @staticmethod
    def from_objs(*args: Tuple[Any], **kwargs: Any) -> NodeCredentials:
        # TODO: we should investigate a way to automatically coerce the right types
        # back and forth with Pydantic and our storage layer
        return NodeCredentials(
            node_uid=kwargs["node_uid"].no_dash,  # type: ignore
            node_name=kwargs["node_name"],
            node_type=kwargs["node_type"],
            verify_key=kwargs["verify_key"].encode(encoder=HexEncoder).decode("utf-8"),  # type: ignore
        )

    def validate(self, key: Union[str, VerifyKey]) -> bool:
        if isinstance(key, VerifyKey):
            verify_key = VerifyKey(self.verify_key, encoder=HexEncoder)
        else:
            verify_key = self.verify_key

        if verify_key == key:
            return True

        raise InvalidNodeCredentials(
            f"Credentials for {self.node_name}: {self.node_uid} do not match"
        )



# stdlib
from typing import Optional

# third party
from nacl.signing import VerifyKey

# relative
from ...core.node.abstract.node_service_interface import NodeServiceInterface
from ...core.node.common.node_service.auth import service_auth
from ...core.node.common.node_service.generic_payload.syft_message import (
    NewSyftMessage as SyftMessage,
)
from ...core.node.common.node_service.node_service import NodeService
from .registry import VMMessageRegistry


class VMServiceClass(NodeService):
    @staticmethod
    @service_auth(guests_welcome=True)  # Service level authentication
    def process(
        node: NodeServiceInterface,
        msg: SyftMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> SyftMessage:
        """A single service to execute messages for the domain client.

        Args:
            node (NodeServiceInterface): domain service interface.
            msg (NewSyftMessage): the message that needs to executed.
            verify_key (Optional[VerifyKey], optional): unique verification of the user. Defaults to None.

        Returns:
            SyftMessage: response message.

        Note:
            TODO: Error and Exception Handling. Errors can be messages of `NewSyftMessage` type.
            Instead of holding the result, they will hold the errors. In short an error is a `NewSyftMessage`
            itself and is propagated to the end user. While when an exception is occurred, the flow should
            be broken to prevent any further execution and an exception type message should be raised.
            The error/exception message handles if it can be private or public.
        """
        msg.check_permissions(node=node, verify_key=verify_key)
        result = msg.run(node=node, verify_key=verify_key).dict()  # type: ignore
        payload_class = msg.__class__
        return payload_class(address=msg.reply_to, kwargs=result, reply=True)  # type: ignore

    @staticmethod
    def message_handler_types() -> list:
        registered_messages = VMMessageRegistry().get_registered_messages()
        return registered_messages



# stdlib
from typing import Optional

# relative
from . import functions  # noqa: 401
from ...ast import add_classes
from ...ast import add_methods
from ...ast import add_modules
from ...ast.globals import Globals
from ..node.abstract.node import AbstractNodeClient
from .autodp.gamma_tensor import GammaTensor
from .autodp.phi_tensor import PhiTensor
from .fixed_precision_tensor import FixedPrecisionTensor
from .nn import Model
from .smpc.share_tensor import ShareTensor
from .tensor import Tensor
from .tensor import TensorPointer  # noqa: 401


def create_tensor_ast(client: Optional[AbstractNodeClient] = None) -> Globals:
    ast = Globals(client)

    modules = [
        "syft",
        "syft.core",
        "syft.core.tensor",
        "syft.core.tensor.tensor",
        "syft.core.tensor.smpc",
        "syft.core.tensor.smpc.share_tensor",
        "syft.core.tensor.fixed_precision_tensor",
        "syft.core.tensor.autodp",
        "syft.core.tensor.autodp.phi_tensor",
        "syft.core.tensor.autodp.gamma_tensor",
        "syft.core.tensor.nn",
    ]
    classes = [
        ("syft.core.tensor.tensor.Tensor", "syft.core.tensor.tensor.Tensor", Tensor),
        (
            "syft.core.tensor.autodp.phi_tensor.PhiTensor",
            "syft.core.tensor.autodp.phi_tensor.PhiTensor",
            PhiTensor,
        ),
        (
            "syft.core.tensor.autodp.gamma_tensor.GammaTensor",
            "syft.core.tensor.autodp.gamma_tensor.GammaTensor",
            GammaTensor,
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
            ShareTensor,
        ),
        (
            "syft.core.tensor.fixed_precision_tensor.FixedPrecisionTensor",
            "syft.core.tensor.fixed_precision_tensor.FixedPrecisionTensor",
            FixedPrecisionTensor,
        ),
        (
            "syft.core.tensor.nn.Model",
            "syft.core.tensor.nn.Model",
            Model,
        ),
    ]

    methods = [
        # # Tensor
        ("syft.core.tensor.tensor.Tensor.T", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.__abs__", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.__add__", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.__divmod__", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.__eq__", "syft.core.tensor.tensor.Tensor"),
        (
            "syft.core.tensor.tensor.Tensor.__floordiv__",
            "syft.core.tensor.tensor.Tensor",
        ),
        ("syft.core.tensor.tensor.Tensor.__ge__", "syft.core.tensor.tensor.Tensor"),
        (
            "syft.core.tensor.tensor.Tensor.__getitem__",
            "syft.core.tensor.tensor.Tensor",
        ),
        ("syft.core.tensor.tensor.Tensor.__gt__", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.__index__", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.__invert__", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.__le__", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.__len__", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.__lshift__", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.__lt__", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.__matmul__", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.__mul__", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.__ne__", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.__neg__", "syft.core.tensor.tensor.Tensor"),
        (
            "syft.core.tensor.tensor.Tensor.__pos__",
            "syft.core.tensor.tensor.Tensor",
        ),  # useless?
        ("syft.core.tensor.tensor.Tensor.__pow__", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.__radd__", "syft.core.tensor.tensor.Tensor"),
        (
            "syft.core.tensor.tensor.Tensor.__rdivmod__",
            "syft.core.tensor.tensor.Tensor",
        ),
        (
            "syft.core.tensor.tensor.Tensor.__rfloordiv__",
            "syft.core.tensor.tensor.Tensor",
        ),
        (
            "syft.core.tensor.tensor.Tensor.__rlshift__",
            "syft.core.tensor.tensor.Tensor",
        ),
        (
            "syft.core.tensor.tensor.Tensor.__rmatmul__",
            "syft.core.tensor.tensor.Tensor",
        ),
        ("syft.core.tensor.tensor.Tensor.__rmul__", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.__rpow__", "syft.core.tensor.tensor.Tensor"),
        (
            "syft.core.tensor.tensor.Tensor.__rrshift__",
            "syft.core.tensor.tensor.Tensor",
        ),
        ("syft.core.tensor.tensor.Tensor.__rshift__", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.__rsub__", "syft.core.tensor.tensor.Tensor"),
        (
            "syft.core.tensor.tensor.Tensor.__rtruediv__",
            "syft.core.tensor.tensor.Tensor",
        ),
        ("syft.core.tensor.tensor.Tensor.__sub__", "syft.core.tensor.tensor.Tensor"),
        (
            "syft.core.tensor.tensor.Tensor.__truediv__",
            "syft.core.tensor.tensor.Tensor",
        ),
        ("syft.core.tensor.tensor.Tensor.argmax", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.argmin", "syft.core.tensor.tensor.Tensor"),
        # ("syft.core.tensor.tensor.Tensor.argsort", "syft.core.tensor.tensor.Tensor"),
        # # ("syft.core.tensor.tensor.Tensor.backward", "syft.lib.python.Bool"),
        # # ("syft.core.tensor.tensor.Tensor.child", "syft.core.tensor.tensor.Tensor"),  # obj level
        ("syft.core.tensor.tensor.Tensor.clip", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.copy", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.cumprod", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.cumsum", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.diagonal", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.dot", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.flatten", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.ravel", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.compress", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.swapaxes", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.gamma", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.max", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.mean", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.min", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.ndim", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.private", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.prod", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.repeat", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.reshape", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.resize", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.shape", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.sort", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.squeeze", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.std", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.sum", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.take", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.tag", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.transpose", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.__pos__", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.put", "syft.core.tensor.tensor.Tensor"),
        ("syft.core.tensor.tensor.Tensor.trace", "syft.core.tensor.tensor.Tensor"),
        (
            "syft.core.tensor.tensor.Tensor.bit_decomposition",
            "syft.lib.python._SyNone",
        ),
        ("syft.core.tensor.tensor.Tensor.mpc_swap", "syft.core.tensor.tensor.Tensor"),
        (
            "syft.core.tensor.tensor.Tensor.concatenate",
            "syft.core.tensor.tensor.Tensor",
        ),
        (
            "syft.core.tensor.tensor.Tensor.exp",
            "syft.core.tensor.tensor.Tensor",
        ),
        (
            "syft.core.tensor.tensor.Tensor.reciprocal",
            "syft.core.tensor.tensor.Tensor",
        ),
        (
            "syft.core.tensor.tensor.Tensor.softmax",
            "syft.core.tensor.tensor.Tensor",
        ),
        (
            "syft.core.tensor.tensor.Tensor.one_hot",
            "syft.core.tensor.tensor.Tensor",
        ),
        (
            "syft.core.tensor.tensor.Tensor.ones_like",
            "syft.core.tensor.tensor.Tensor",
        ),
        # # SMPC
        # (
        #     "syft.core.tensor.tensor.Tensor.fix_precision",
        #     "syft.core.tensor.tensor.Tensor",
        # ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.generate_przs",
            "syft.core.tensor.tensor.Tensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.generate_przs_on_dp_tensor",
            "syft.core.tensor.tensor.Tensor",
        ),
        # ("syft.core.tensor.tensor.Tensor.share", "syft.core.tensor.tensor.Tensor"),
        # Share Tensor Operations
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.__add__",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.__sub__",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.__mul__",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.__gt__",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.__matmul__",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.__rmatmul__",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.sum",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.repeat",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.copy",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.diagonal",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.flatten",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.transpose",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.resize",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.ravel",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.compress",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.reshape",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.squeeze",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.swapaxes",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.__pos__",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.put",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.__neg__",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.take",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.cumsum",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.trace",
            "syft.core.tensor.smpc.share_tensor.ShareTensor",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.crypto_store",
            "syft.core.smpc.store.CryptoStore",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.populate_store",
            "syft.lib.python._SyNone",
        ),
        (
            "syft.core.tensor.smpc.share_tensor.ShareTensor.bit_decomposition",
            "syft.lib.python._SyNone",
        ),
        # nn Modules
        (
            "syft.core.tensor.nn.Model.fit",
            "syft.lib.python._SyNone",
        ),
        (
            "syft.core.tensor.nn.Model.step",
            "syft.lib.python._SyNone",
        ),
    ]

    add_modules(ast, modules)
    add_classes(ast, classes)
    add_methods(ast, methods)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast



# stdlib
from typing import List
from typing import Type

# third party
from nacl.signing import VerifyKey

# relative
from ......logger import debug
from ......logger import error
from .....store.storeable_object import StorableObject
from ....abstract.node import AbstractNode
from ..auth import service_auth
from ..heritage_update.heritage_update_messages import HeritageUpdateMessage
from ..node_service import ImmediateNodeServiceWithoutReply
from .child_node_lifecycle_messages import RegisterChildNodeMessage


class ChildNodeLifecycleService(ImmediateNodeServiceWithoutReply):
    @classmethod
    def icon(cls) -> str:
        return "💾"

    @staticmethod
    @service_auth(root_only=True)
    def process(
        node: AbstractNode, msg: RegisterChildNodeMessage, verify_key: VerifyKey
    ) -> None:
        debug(
            f"> Executing {ChildNodeLifecycleService.pprint()} {msg.pprint} on {node.pprint}"
        )
        addr = msg.child_node_client_address
        lookup_id = msg.lookup_id  # TODO: Fix, see above

        node.store[lookup_id] = StorableObject(id=lookup_id, data=addr)

        debug(
            (
                f"> Saving 💾 {addr.pprint} {addr.target_emoji()} with "
                + f"Key: {lookup_id} ➡️ {type(node.store)}"
            )
        )

        # Step 2: update the child node and its descendants with our node.id in their
        # .address objects

        # WARNING: This has no where to go
        # ---------------------------------
        # the goal of this code is to tell the child_node_client, hey you need to update
        # yourself with this new Address (which is up to 4 locations)
        # the old code didn't serialize the incoming object which was actually a real
        # python pointer to the original client
        # now that its a serialized address there are no pointers in memory to the
        # original child clients send_immediate_msg_without_reply function so
        # there is no way to invoke it
        debug(f"> Sending 👪 Update from {node.pprint} back to {addr.target_emoji()}")
        debug("> Update Contains", type(node.address), node.address)
        heritage_msg = HeritageUpdateMessage(
            new_ancestry_address=node.address, address=msg.child_node_client_address
        )

        location = msg.child_node_client_address.target_id.id
        try:
            in_memory_client = node.in_memory_client_registry[location]
            # we need to sign here with the current node not the destination side
            in_memory_client.send_immediate_msg_without_reply(msg=heritage_msg)
            debug(f"> Forwarding {msg.pprint} to {addr.target_emoji()}")
            return None
        except Exception as e:
            error(f"{location} not on nodes in_memory_client. {e}")
            pass
        return None

    @staticmethod
    def message_handler_types() -> List[Type[RegisterChildNodeMessage]]:
        return [RegisterChildNodeMessage]



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Set
from typing import Tuple
from typing import Union

# third party
import numpy as np
from numpy.typing import ArrayLike
import pandas as pd

# relative
from ..common.serde.serializable import serializable
from .data_subject import DataSubject


# allow us to serialize and deserialize np.arrays with strings inside as two np.arrays
# one containing the uint8 bytes and the other the offsets between strings
# TODO: Should move to a vectorized version.
def numpyutf8tolist(string_index: Tuple[np.ndarray, np.ndarray]) -> np.ndarray:
    index_length = int(string_index[-1])
    index_array = string_index[-(index_length + 1) : -1]  # noqa
    string_array: np.ndarray = string_index[: -(index_length + 1)]
    output_bytes: bytes = string_array.astype(np.uint8).tobytes()
    output_list = []
    last_offset = 0
    for offset in index_array:
        chars = output_bytes[last_offset:offset]
        final_string = chars.decode("utf-8")
        last_offset = offset
        output_list.append(final_string)
    return np.array(output_list)


def liststrtonumpyutf8(string_list: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    bytes_list = []
    indexes = []
    offset = 0
    for item in string_list:
        if not isinstance(item, (DataSubject, str)):
            raise Exception(
                f"DataSubjectList entities must be List[Union[str, DataSubject]]. {type(item)}"
            )
        name = item if isinstance(item, str) else item.name
        name_bytes = name.encode("utf-8")
        offset += len(name_bytes)
        indexes.append(offset)
        bytes_list.append(name_bytes)

    np_bytes = np.frombuffer(b"".join(bytes_list), dtype=np.uint8)
    np_bytes = np_bytes.astype(np.uint64)
    np_indexes = np.array(indexes, dtype=np.uint64)
    index_length = np.array([len(np_indexes)], dtype=np.uint64)
    output_array = np.concatenate([np_bytes, np_indexes, index_length])
    return output_array


@serializable(recursive_serde=True)
class DataSubjectList:
    __attr_allowlist__ = ("one_hot_lookup", "data_subjects_indexed")
    __slots__ = ("one_hot_lookup", "data_subjects_indexed")

    # one_hot_lookup is a numpy array of unicode strings which can't be serialized
    __serde_overrides__ = {
        "one_hot_lookup": [liststrtonumpyutf8, numpyutf8tolist],
    }

    def __init__(
        self,
        one_hot_lookup: np.ndarray[Union[DataSubject, str, np.integer]],
        data_subjects_indexed: np.ndaray,
    ) -> None:
        self.one_hot_lookup = one_hot_lookup
        self.data_subjects_indexed = data_subjects_indexed

    @staticmethod
    def from_series(entities_dataframe_slice: pd.Series) -> DataSubjectList:
        """Given a Pandas Series object (such as from
        getting a column from a pandas DataFrame, return an DataSubjectList"""

        # This will be the equivalent of the DataSubjectList.data_subjects_indexed
        if not isinstance(entities_dataframe_slice, np.ndarray):
            data_subjects = entities_dataframe_slice.to_numpy()
        else:
            data_subjects = entities_dataframe_slice

        # data_subjects = (
        #     entities_dataframe_slice.to_numpy()
        #     if not isinstance(entities_dataframe_slice, np.ndarray)
        #     else entities_dataframe_slice
        # )
        unique_data_subjects, data_subjects_indexed = np.unique(
            data_subjects, return_inverse=True
        )

        # This will be the equivalent of the DataSubjectList.one_hot_indexed- a sorted array of all unique data_subjects
        unique_data_subjects = unique_data_subjects.astype(np.str_)

        return DataSubjectList(
            one_hot_lookup=unique_data_subjects,
            data_subjects_indexed=data_subjects_indexed,
        )

    @staticmethod
    def from_objs(entities: Union[np.ndarray, list]) -> DataSubjectList:

        entities = np.array(entities, copy=False)
        one_hot_lookup, entities_indexed = np.unique(entities, return_inverse=True)
        if entities_indexed.shape != entities.shape:
            entities_indexed.resize(entities.shape)
        return DataSubjectList(one_hot_lookup, entities_indexed)

    # def __getitem__(self, key: Union[int, slice, str]) -> Union[DataSubject, str]:
    #     return self.one_hot_lookup[self.data_subjects_indexed[key]]

    def copy(self) -> DataSubjectList:
        return DataSubjectList(
            self.one_hot_lookup.copy(), self.data_subjects_indexed.copy()
        )

    def __len__(self) -> int:
        return len(self.data_subjects_indexed)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, DataSubjectList):
            dsi_comparison = self.data_subjects_indexed == other.data_subjects_indexed
            ohl_comparison = self.one_hot_lookup == other.one_hot_lookup
            ohl_comparison = (
                ohl_comparison
                if isinstance(ohl_comparison, bool)
                else ohl_comparison.all()
            )
            if isinstance(dsi_comparison, bool):
                if dsi_comparison is True and ohl_comparison is True:
                    return True
                else:
                    return False
            elif isinstance(dsi_comparison, np.ndarray):
                if dsi_comparison.all() is True and ohl_comparison is True:
                    return True
                else:
                    return False
        return self == other

    def sum(self, target_shape: tuple) -> DataSubjectList:
        """
        ::target_shape:: is the shape that the data (DP Tensor.child) has.
        """
        return DataSubjectList(
            self.one_hot_lookup.copy(),
            self.data_subjects_indexed.reshape((-1, *target_shape)),
        )

    def flatten(self) -> DataSubjectList:
        return DataSubjectList(
            one_hot_lookup=self.one_hot_lookup.copy(),
            data_subjects_indexed=self.data_subjects_indexed.flatten(),
        )

    def transpose(self: DataSubjectList, axes: Tuple) -> DataSubjectList:
        return DataSubjectList(
            self.one_hot_lookup, np.transpose(self.data_subjects_indexed, axes=axes)
        )

    @staticmethod
    def combine(dsl1: DataSubjectList, dsl2: DataSubjectList) -> DataSubjectList:

        """
        From Ishan's PR: https://github.com/OpenMined/PySyft/pull/6490/

        This code lets us combine two Data Subject Lists.
        This is essential when going from PhiTensors -> GammaTensors.

        The benefit of this code over recalling np.unique(return_indexed) is that we don't
        iterate over the entire 2nd data subject list.

        :param dsl1:
        :param dsl2:
        :return:
        """

        if dsl1.one_hot_lookup.size == 0:
            return dsl2
        elif dsl2.one_hot_lookup.size == 0:
            return dsl1
        else:

            # dsl1_uniques = dsl1.num_uniques
            # dsl2_uniques = dsl2.num_uniques
            dsl1_uniques = len(dsl1.one_hot_lookup)
            dsl2_uniques = len(dsl2.one_hot_lookup)

            if dsl1_uniques >= dsl2_uniques:
                search_terms = dsl2.one_hot_lookup
                bigger_list = dsl1.one_hot_lookup
                array_to_change = dsl2.data_subjects_indexed
                unchanged_array = dsl1.data_subjects_indexed
            else:
                search_terms = dsl1.one_hot_lookup
                bigger_list = dsl2.one_hot_lookup
                array_to_change = dsl1.data_subjects_indexed
                unchanged_array = dsl2.data_subjects_indexed

            # Array of True/False depending on whether search term exists in bigger list
            overlap = np.isin(
                search_terms, bigger_list
            )  # TODO: is there a way to use np.searchsorted w/o the index 0 problem?

            # The DS at these indices of search_terms exist in bigger_list
            search = overlap == True  # noqa: E712
            overlapping_indices = search.nonzero()[0]
            # The DS at these indices are unique
            unique_indices = np.invert(search).nonzero()[0]  # noqa: E712

            # Suppressing E712 above because the recommended way (is True) does not work elementwise

            if (
                len(overlapping_indices) == 0
            ):  # If there's no overlap, our job is super simple
                return DataSubjectList(
                    one_hot_lookup=np.concatenate(
                        (dsl1.one_hot_lookup, dsl2.one_hot_lookup)
                    ),
                    data_subjects_indexed=np.stack(
                        (
                            dsl1.data_subjects_indexed,
                            dsl2.data_subjects_indexed + dsl1_uniques,
                        )
                    ),
                )

            # Task 1- For the overlapping data subjects, we need to find the index already allotted to them.
            target_overlap_indices = np.searchsorted(
                bigger_list, search_terms.take(overlapping_indices)
            )

            # Now that we need to replace the previous indices with the new indices
            output_data_subjects_indexed = np.zeros_like(dsl2.data_subjects_indexed)
            for old_value, new_value in zip(
                overlapping_indices, target_overlap_indices
            ):
                output_data_subjects_indexed[array_to_change == old_value] = new_value

            # Task 2- do the same but for unique data subjects
            unique_data_subjects = search_terms.take(unique_indices)

            output_one_hot_encoding = np.concatenate(
                (bigger_list, unique_data_subjects)
            )
            target_unique_indices = np.arange(
                len(bigger_list), len(output_one_hot_encoding)
            )

            for old_value, new_value in zip(unique_indices, target_unique_indices):
                output_data_subjects_indexed[array_to_change == old_value] = new_value

            final_dsi = np.stack((unchanged_array, output_data_subjects_indexed))

            return DataSubjectList(
                one_hot_lookup=output_one_hot_encoding, data_subjects_indexed=final_dsi
            )

    @staticmethod
    def absorb(dsl1: DataSubjectList, dsl2: DataSubjectList) -> DataSubjectList:
        # TODO: Check if DataSubjectArrays need to be flattened before appending, if different sizes?

        """
        Unlike Combine() which creates a DataSubjectArray for a GammaTensor where one data point is owned
        by multiple data subjects, Absorb() creates a GammaTensor where each data point is
        still only owned by a single data subject.


        :param dsl1:
        :param dsl2:
        :return:
        """

        if dsl1.one_hot_lookup.size == 0:
            return dsl2
        elif dsl2.one_hot_lookup.size == 0:
            return dsl1
        else:

            # dsl1_uniques = dsl1.num_uniques
            # dsl2_uniques = dsl2.num_uniques
            dsl1_uniques = len(dsl1.one_hot_lookup)
            dsl2_uniques = len(dsl2.one_hot_lookup)

            if dsl1_uniques >= dsl2_uniques:
                search_terms = dsl2.one_hot_lookup
                bigger_list = dsl1.one_hot_lookup
                array_to_change = dsl2.data_subjects_indexed
                unchanged_array = dsl1.data_subjects_indexed
            else:
                search_terms = dsl1.one_hot_lookup
                bigger_list = dsl2.one_hot_lookup
                array_to_change = dsl1.data_subjects_indexed
                unchanged_array = dsl2.data_subjects_indexed

            # Array of True/False depending on whether search term exists in bigger list
            overlap = np.isin(
                search_terms, bigger_list
            )  # TODO: is there a way to use np.searchsorted w/o the index 0 problem?

            # The DS at these indices of search_terms exist in bigger_list
            search = overlap == True  # noqa: E712
            overlapping_indices = search.nonzero()[0]
            # The DS at these indices are unique
            unique_indices = np.invert(search).nonzero()[0]

            # Suppressing E712 above because the recommended way (is True) does not work elementwise

            if (
                len(overlapping_indices) == 0
            ):  # If there's no overlap, our job is super simple
                return DataSubjectList(
                    one_hot_lookup=np.concatenate(
                        (dsl1.one_hot_lookup, dsl2.one_hot_lookup)
                    ),
                    data_subjects_indexed=np.append(  # Should this be concatenate?
                        dsl1.data_subjects_indexed,
                        dsl2.data_subjects_indexed + dsl1_uniques,
                    ),
                )

            # Task 1- For the overlapping data subjects, we need to find the index already allotted to them.
            target_overlap_indices = np.searchsorted(
                bigger_list, search_terms.take(overlapping_indices)
            )

            # Now that we need to replace the previous indices with the new indices
            output_data_subjects_indexed = np.zeros_like(dsl2.data_subjects_indexed)
            for old_value, new_value in zip(
                overlapping_indices, target_overlap_indices
            ):
                output_data_subjects_indexed[array_to_change == old_value] = new_value

            # Task 2- do the same but for unique data subjects
            unique_data_subjects = search_terms.take(unique_indices)

            output_one_hot_encoding = np.concatenate(
                (bigger_list, unique_data_subjects)
            )
            target_unique_indices = np.arange(
                len(bigger_list), len(output_one_hot_encoding)
            )

            for old_value, new_value in zip(unique_indices, target_unique_indices):
                output_data_subjects_indexed[array_to_change == old_value] = new_value

            final_dsi = np.append(unchanged_array, output_data_subjects_indexed)

            return DataSubjectList(
                one_hot_lookup=output_one_hot_encoding, data_subjects_indexed=final_dsi
            )


def numpyutf8todslarray(input_index: Tuple[np.ndarray, np.ndarray]) -> ArrayLike:
    """Decodes utf-8 encoded numpy array to DataSubjectArray.

    Args:
        string_index (Tuple[np.ndarray, np.ndarray]): encoded array

    Returns:
        np.ndarray: decoded DataSubjectArray.
    """
    shape_length = int(input_index[-1])
    shape = tuple(input_index[-(shape_length + 1) : -1])  # noqa
    string_index = input_index[: -(shape_length + 1)]
    index_length = int(string_index[-1])
    index_array = string_index[-(index_length + 1) : -1]  # noqa
    string_array: np.ndarray = string_index[: -(index_length + 1)]
    output_bytes: bytes = string_array.astype(np.uint8).tobytes()
    output_list = []
    last_offset = 0
    for offset in index_array:
        chars = output_bytes[last_offset:offset]
        final_string = DataSubjectArray.fromstring(chars.decode("utf-8"))
        last_offset = offset
        output_list.append(final_string)
    return np.array(output_list).reshape(shape)


def dslarraytonumpyutf8(string_list: np.ndarray) -> ArrayLike:
    """Encodes DataSubjectArray to utf-8 encoded numpy array.

    Args:
        string_list (np.ndarray): DataSubjectArray to be encoded

    Raises:
        Exception: DataSubject is not a DataSubjectArray

    Returns:
        Tuple[np.ndarray, np.ndarray]: utf-8 encoded int Numpy array
    """
    # print("dsl list before", string_list)
    array_shape = string_list.shape
    string_list = string_list.flatten()
    bytes_list = []
    indexes = []
    offset = 0
    # print("dsl list ", string_list)
    for item in string_list:
        if not isinstance(item, DataSubjectArray):
            raise Exception(
                f"DataSubjectList entities must be  DataSubject. {type(item)}"
            )
        name = item.tostring()
        name_bytes = name.encode("utf-8")
        offset += len(name_bytes)
        indexes.append(offset)
        bytes_list.append(name_bytes)

    np_bytes = np.frombuffer(b"".join(bytes_list), dtype=np.uint8)
    np_bytes = np_bytes.astype(np.uint64)
    np_indexes = np.array(indexes, dtype=np.uint64)
    index_length = np.array([len(np_indexes)], dtype=np.uint64)
    shape = np.array(array_shape, dtype=np.uint64)
    shape_length = np.array([len(shape)], dtype=np.uint64)
    output_array = np.concatenate(
        [np_bytes, np_indexes, index_length, shape, shape_length]
    )

    return output_array


@serializable(recursive_serde=True)
class DataSubjectArray:
    __attr_allowlist__ = ("data_subjects",)

    delimiter = ","

    def __init__(self, data_subjects: Union[str, List[str], Set[str]]):
        self.data_subjects = set(data_subjects)

    def __len__(self) -> int:
        return len(self.data_subjects)

    def tostring(self) -> str:
        return f"{self.delimiter}".join(self.data_subjects)

    @classmethod
    def fromstring(cls, input_string: str) -> DataSubjectArray:
        return DataSubjectArray(set(input_string.split(f"{cls.delimiter}")))

    def __add__(self, other: Union[DataSubjectArray, Any]) -> DataSubjectArray:
        if isinstance(other, DataSubjectArray):
            return DataSubjectArray(self.data_subjects.union(other.data_subjects))
        else:
            return DataSubjectArray(self.data_subjects)

    def __sub__(self, other: Union[DataSubjectArray, Any]) -> DataSubjectArray:
        if isinstance(other, DataSubjectArray):
            return DataSubjectArray(self.data_subjects.union(other.data_subjects))
        else:
            return DataSubjectArray(self.data_subjects)

    def __mul__(self, other: Union[DataSubjectArray, Any]) -> DataSubjectArray:
        if isinstance(other, DataSubjectArray):
            return DataSubjectArray(self.data_subjects.union(other.data_subjects))
        else:
            return DataSubjectArray(self.data_subjects)

    def __ge__(self, other: Union[DataSubjectArray, Any]) -> DataSubjectArray:
        if isinstance(other, DataSubjectArray):
            return DataSubjectArray(self.data_subjects.union(other.data_subjects))
        else:
            return DataSubjectArray(self.data_subjects)

    def __le__(self, other: Union[DataSubjectArray, Any]) -> DataSubjectArray:
        if isinstance(other, DataSubjectArray):
            return DataSubjectArray(self.data_subjects.union(other.data_subjects))
        else:
            return DataSubjectArray(self.data_subjects)

    def __gt__(self, other: Union[DataSubjectArray, Any]) -> DataSubjectArray:
        if isinstance(other, DataSubjectArray):
            return DataSubjectArray(self.data_subjects.union(other.data_subjects))
        else:
            return DataSubjectArray(self.data_subjects)

    def exp(self) -> DataSubjectArray:
        return DataSubjectArray(self.data_subjects)

    def __lt__(self, other: Union[DataSubjectArray, Any]) -> DataSubjectArray:
        if isinstance(other, DataSubjectArray):
            return DataSubjectArray(self.data_subjects.union(other.data_subjects))
        else:
            return DataSubjectArray(self.data_subjects)

    def __truediv__(self, other: Union[DataSubjectArray, Any]) -> DataSubjectArray:
        if isinstance(other, DataSubjectArray):
            return DataSubjectArray(self.data_subjects.union(other.data_subjects))
        else:
            return DataSubjectArray(self.data_subjects)

    def __rmatmul__(self, other: Union[DataSubjectArray, Any]) -> DataSubjectArray:
        if isinstance(other, DataSubjectArray):
            return DataSubjectArray(self.data_subjects.union(other.data_subjects))
        else:
            return DataSubjectArray(self.data_subjects)

    def __rtruediv__(self, other: Union[DataSubjectArray, Any]) -> DataSubjectArray:
        if isinstance(other, DataSubjectArray):
            return DataSubjectArray(other.data_subjects.union(self.data_subjects))
        else:
            return DataSubjectArray(self.data_subjects)

    def __repr__(self) -> str:
        return "DataSubjectArray: " + str(self.data_subjects.__repr__())

    def __iter__(self) -> Iterator[Any]:
        for data_subject in self.data_subjects:
            yield data_subject

    def __contains__(self, item: Union[str, DataSubjectArray]) -> bool:
        if isinstance(item, DataSubjectArray):
            return self.data_subjects.isdisjoint(item.data_subjects)
        else:
            return self.data_subjects.isdisjoint(set(item))

    def conjugate(self, *args: List[Any], **kwargs: Dict[Any, Any]) -> DataSubjectArray:
        return DataSubjectArray(self.data_subjects)

    def subtract(
        self,
        x: Union[DataSubjectArray, Any],
        y: Union[DataSubjectArray, Any],
        *args: List[Any],
        **kwargs: Dict[Any, Any],
    ) -> DataSubjectArray:
        if isinstance(y, DataSubjectArray) and isinstance(x, DataSubjectArray):
            return DataSubjectArray(x.data_subjects.union(y.data_subjects))
        elif isinstance(y, DataSubjectArray):
            return DataSubjectArray(y.data_subjects)
        elif isinstance(x, DataSubjectArray):
            return DataSubjectArray(x.data_subjects)
        else:
            raise ValueError(
                f"Either X:{type(x)} Y:{type(y)} should be DataSubjectArray"
            )

    def multiply(
        self,
        x: Union[DataSubjectArray, Any],
        y: Union[DataSubjectArray, Any],
        *args: List[Any],
        **kwargs: Dict[Any, Any],
    ) -> DataSubjectArray:
        if isinstance(y, DataSubjectArray) and isinstance(x, DataSubjectArray):
            return DataSubjectArray(x.data_subjects.union(y.data_subjects))
        elif isinstance(y, DataSubjectArray):
            return DataSubjectArray(y.data_subjects)
        elif isinstance(x, DataSubjectArray):
            return DataSubjectArray(x.data_subjects)
        else:
            raise ValueError(
                f"Either X:{type(x)} Y:{type(y)} should be DataSubjectArray"
            )

    def __eq__(self, other: Union[DataSubjectArray, Any]) -> bool:
        if isinstance(other, DataSubjectArray):
            return self.data_subjects == other.data_subjects
        elif isinstance(other, np.ndarray):
            return np.array(self) == other
        else:
            raise NotImplementedError

    def log(self) -> DataSubjectArray:
        return DataSubjectArray(self.data_subjects)

    def real(self) -> DataSubjectArray:
        return DataSubjectArray(self.data_subjects)

    def var(self, *args: List[Any], **kwargs: Dict[Any, Any]) -> DataSubjectArray:
        return (self - np.mean(self)) * (self - np.mean(self))

    def sqrt(self, *args: List[Any], **kwargs: Dict[Any, Any]) -> DataSubjectArray:
        return DataSubjectArray(self.data_subjects)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs) -> ArrayLike:  # type: ignore
        method_name = ufunc.__name__
        print("method_name", method_name)
        method = getattr(self, method_name, None)
        if method is not None:
            return method(*inputs, **kwargs)
        else:
            raise NotImplementedError(
                f"Method: {method_name} not implemented in DataSubjectArray"
            )

    @staticmethod
    def from_objs(input_subjects: Union[np.ndarray, list]) -> ArrayLike:
        # TODO: When the user passes the data subjects they might pass it as list
        # specifying the entity per row, but in our new notation we want it to be
        # per data point, we should make sure that we implement in such a way we expand
        # the datasubjects automatically for row to data point mapping.
        if not isinstance(input_subjects, np.ndarray):
            input_subjects = np.array(input_subjects)

        data_map = (
            lambda x: DataSubjectArray([str(x)])
            if not isinstance(x, DataSubjectArray)
            else x
        )
        map_function = np.vectorize(data_map)

        return map_function(input_subjects)



""" A set of useful methods used by the syft.core.node.common.node_table submodule. """
# stdlib
from typing import Any
from typing import Dict
from typing import Tuple as TypeTuple

# third party
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session
from sqlalchemy.orm import sessionmaker

# relative
from . import Base
from .roles import Role

# attribute names representing a date owned by the PyGrid's database schemas.
datetime_cols = ["date", "created_at", "destroyed_at", "deployed_on", "updated_on"]


def model_to_json(model: Base) -> Dict[str, Any]:
    """
    Returns a JSON representation of an SQLAlchemy-backed object.

    Args:
        model: SQLAlchemy-backed object to be represented as a JSON data structure.
    Returns:
        Dict: Python dictionary representing the SQLAlchemy object.
    """
    json = {}
    for col in model.__mapper__.attrs.keys():  # type: ignore
        if col != "hashed_password" and col != "salt":
            if col in datetime_cols:
                # Cast datetime object to string
                json[col] = str(getattr(model, col))
            else:
                json[col] = getattr(model, col)
    return json


def seed_db(db: Session) -> None:
    """Creates the standard PyGrid roles and seeds it into the database.

    Args:
        db (Session): Database Session object used to handle database operations.
    """
    new_role = Role(
        name="Data Scientist",
        can_make_data_requests=True,
        can_triage_data_requests=False,
        can_manage_privacy_budget=False,
        can_create_users=False,
        can_manage_users=False,
        can_edit_roles=False,
        can_manage_infrastructure=False,
        can_upload_data=False,
        can_upload_legal_document=False,
        can_edit_domain_settings=False,
    )
    db.add(new_role)

    new_role = Role(
        name="Compliance Officer",
        can_make_data_requests=False,
        can_triage_data_requests=True,
        can_manage_privacy_budget=False,
        can_create_users=False,
        can_manage_users=True,
        can_edit_roles=False,
        can_manage_infrastructure=False,
        can_upload_data=False,
        can_upload_legal_document=False,
        can_edit_domain_settings=False,
    )
    db.add(new_role)

    new_role = Role(
        name="Administrator",
        can_make_data_requests=True,
        can_triage_data_requests=True,
        can_manage_privacy_budget=True,
        can_create_users=True,
        can_manage_users=True,
        can_edit_roles=True,
        can_manage_infrastructure=True,
        can_upload_data=True,
        can_upload_legal_document=True,
        can_edit_domain_settings=True,
    )
    db.add(new_role)

    new_role = Role(
        name="Owner",
        can_make_data_requests=True,
        can_triage_data_requests=True,
        can_manage_privacy_budget=True,
        can_create_users=True,
        can_manage_users=True,
        can_edit_roles=True,
        can_manage_infrastructure=True,
        can_upload_data=True,
        can_upload_legal_document=True,
        can_edit_domain_settings=True,
    )
    db.add(new_role)
    db.commit()


def create_memory_db_engine() -> TypeTuple[Engine, sessionmaker]:
    """Creates the database tables in memory instead of connecting with a real database server.

    Returns:
        Tuple: Tuple containing both database engine and session maker.
    """
    db_engine = create_engine("sqlite://", echo=False)
    Base.metadata.create_all(db_engine)  # type: ignore
    SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=db_engine)
    seed_db(SessionLocal())
    return db_engine, SessionLocal



# stdlib
from typing import Union

# relative
from .....grid import GridURL
from ...abstract.node import AbstractNodeClient

ClientLike = Union[AbstractNodeClient, GridURL]


class NewRequestAPI:
    def __init__(self, client: AbstractNodeClient):
        self.client = client

    def get_client(self, client: ClientLike) -> AbstractNodeClient:
        if isinstance(client, AbstractNodeClient):
            client.signing_key = self.client.signing_key
            client.verify_key = self.client.signing_key.verify_key
            return client

        if isinstance(client, str):
            client = GridURL.from_url(client)

        # relative
        from .....grid.client.client import connect

        return connect(
            url=client.with_path("/api/v1"),
            timeout=10,
            user_key=self.client.signing_key,
        )

    def get_client_url(self, client: ClientLike) -> GridURL:
        if isinstance(client, (str, GridURL)):
            return GridURL.from_url(client)
        else:
            return client.routes[0].connection.base_url  # type: ignore



# stdlib
from typing import Any
from typing import Optional
from typing import Tuple
from typing import Union

# third party
import numpy as np
from numpy.typing import NDArray

# relative
from ....common.serde.serializable import serializable
from ...autodp.gamma_tensor import GammaTensor
from ...autodp.phi_tensor import PhiTensor
from ..utils import col2im_indices
from ..utils import im2col_indices
from .base import Layer


@serializable(recursive_serde=True)
class AvgPool(Layer):
    """Average pooling operation for spatial data.
    Parameters
    ----------
    pool_size : tuple of 2 integers,
        factors by which to downscale (vertical, horizontal).
        (2, 2) will halve the image in each dimension.
    Returns
    -------
    4D numpy.array
        with shape `(nb_samples, channels, pooled_rows, pooled_cols)` if dim_ordering='th'
        or 4D tensor with shape:
        `(samples, pooled_rows, pooled_cols, channels)` if dim_ordering='tf'.
    """

    __attr_allowlist__ = (
        "stride",
        "pool_size",
        "input_shape",
        "out_shape",
        "X_col",
    )

    def __init__(self, pool_size: Union[Tuple[int, ...], int], stride: int = 1) -> None:
        if isinstance(pool_size, tuple):
            self.pool_size = pool_size
        elif isinstance(pool_size, int):
            self.pool_size = pool_size, pool_size
        else:
            raise TypeError("Pool size can be either int or Tuple")

        self.stride = stride

        self.input_shape: Optional[Tuple[int, ...]] = None
        self.out_shape: Optional[Tuple[int, ...]] = None

    def connect_to(self, prev_layer: Layer) -> None:

        if not (5 > len(prev_layer.out_shape) >= 3):  # type: ignore
            raise ValueError(
                f"Input layer shape:{prev_layer.out_shape} dimension should be between [3,5)"  # type: ignore
            )

        old_h, old_w = prev_layer.out_shape[-2:]  # type: ignore
        pool_h, pool_w = self.pool_size

        new_h = (old_h - pool_h) // self.stride + 1
        new_w = (old_w - pool_w) // self.stride + 1

        self.out_shape = prev_layer.out_shape[:-2] + (new_h, new_w)  # type: ignore

    def forward(
        self, input: Union[PhiTensor, GammaTensor], *args: Any, **kwargs: Any
    ) -> Union[GammaTensor, PhiTensor]:
        # shape
        self.input_shape = input.shape
        pool_h, pool_w = self.pool_size

        if self.out_shape is None:
            raise ValueError(
                "`self.out_shape` is None. Please check the out_shape is correctly initialized."
            )

        h_out, w_out = self.out_shape[-2:]

        # forward

        n, d, h, w = input.shape
        input_reshaped = input.reshape((n * d, 1, h, w))

        self.X_col = im2col_indices(
            input_reshaped, pool_h, pool_w, padding=0, stride=self.stride
        )
        outputs = self.X_col.mean(axis=0)
        outputs = outputs.reshape((h_out, w_out, n, d))
        outputs = outputs.transpose((2, 3, 0, 1))
        # print("Done with AvgPool forward pass")
        return outputs

    def backward(
        self, pre_grad: Union[PhiTensor, GammaTensor], *args: Any, **kwargs: Any
    ) -> Union[GammaTensor, PhiTensor]:

        if self.input_shape is None:
            raise ValueError(
                "Input shape is None. \
                Please check if the input shape is correctly initialized from the prev_layer."
            )

        n, d, w, h = self.input_shape
        pool_h, pool_w = self.pool_size

        dX_col = self.X_col.zeros_like()

        dout_col = pre_grad.transpose(2, 3, 0, 1).ravel()

        dout_col_size = np.prod(dout_col.shape)

        dX_col[:, range(dout_col_size)] = dout_col * (1.0 / dX_col.shape[0])  # type: ignore

        dX = col2im_indices(
            dX_col, (n * d, 1, h, w), pool_h, pool_w, padding=0, stride=self.stride
        )

        dX = dX.reshape(self.input_shape)

        return dX


@serializable(recursive_serde=True)
class MaxPool(Layer):
    """Max pooling operation for spatial data.
    Parameters
    ----------
    pool_size : tuple of 2 integers,
        factors by which to downscale (vertical, horizontal).
        (2, 2) will halve the image in each dimension.
    Returns
    -------
    4D numpy.array
        with shape `(nb_samples, channels, pooled_rows, pooled_cols)` if dim_ordering='th'
        or 4D tensor with shape:
        `(samples, pooled_rows, pooled_cols, channels)` if dim_ordering='tf'.
    """

    __attr_allowlist__ = (
        "stride",
        "X_col",
        "max_idx",
        "pool_size",
        "input_shape",
        "out_shape",
    )

    def __init__(self, pool_size: Union[Tuple[int, ...], int], stride: int = 1) -> None:
        if isinstance(pool_size, tuple):
            self.pool_size = pool_size
        elif isinstance(pool_size, int):
            self.pool_size = pool_size, pool_size
        else:
            raise TypeError("Pool size can be either int or Tuple")

        self.stride = stride

        self.input_shape: Optional[Tuple[int, ...]] = None
        self.out_shape: Optional[Tuple[int, ...]] = None
        self.X_col: Optional[Union[GammaTensor, PhiTensor]] = None
        self.max_idx: Optional[Union[int, slice, range, NDArray]] = None

    def connect_to(self, prev_layer: Layer) -> None:
        # prev_layer.out_shape: (nb_batch, ..., height, width)
        if len(prev_layer.out_shape) < 3:  # type: ignore
            raise ValueError(
                f"Input layer shape should have at least \
                three dimensions:{prev_layer.out_shape} for MaxPool"  # type: ignore
            )

        old_h, old_w = prev_layer.out_shape[-2:]  # type: ignore
        pool_h, pool_w = self.pool_size
        new_h = (old_h - pool_h) // self.stride + 1
        new_w = (old_w - pool_w) // self.stride + 1

        self.input_shape = prev_layer.out_shape  # type: ignore
        self.out_shape = prev_layer.out_shape[:-2] + (new_h, new_w)  # type: ignore

    def forward(
        self, input: Union[PhiTensor, GammaTensor], *args: Any, **kwargs: Any
    ) -> Union[GammaTensor, PhiTensor]:
        # shape

        self.input_shape = input.shape
        pool_h, pool_w = self.pool_size

        if self.out_shape is None:
            raise ValueError(
                "`self.out_shape` is None. Please check the out_shape is correctly initialized."
            )

        h_out, w_out = self.out_shape[-2:]

        # forward

        n, d, h, w = input.shape
        input_reshaped = input.reshape((n * d, 1, h, w))
        self.X_col = im2col_indices(
            input_reshaped, pool_h, pool_w, padding=0, stride=self.stride
        )

        self.max_idx = self.X_col._argmax(axis=0)
        outputs = self.X_col[self.max_idx, range(self.max_idx.size)]

        outputs = outputs.reshape((h_out, w_out, n, d))

        outputs = outputs.transpose((2, 3, 0, 1))

        return outputs

    def backward(
        self, pre_grad: Union[PhiTensor, GammaTensor], *args: Any, **kwargs: Any
    ) -> Union[GammaTensor, PhiTensor]:

        if self.input_shape is None:
            raise ValueError(
                "Input shape is None. \
                Please check if the input shape is correctly initialized from the prev_layer."
            )

        n, d, w, h = self.input_shape
        pool_h, pool_w = self.pool_size

        if self.X_col is None or self.max_idx is None:
            raise ValueError(
                "X_col or max_idx is not initialized. Please call .forward before calling .backward."
            )

        dX_col = self.X_col.zeros_like()

        dout_col = pre_grad.transpose((2, 3, 0, 1)).ravel()

        dout_col_size = np.prod(dout_col.shape)

        dX_col[self.max_idx, range(dout_col_size)] = dout_col  # type: ignore

        dX = col2im_indices(
            dX_col, (n * d, 1, h, w), pool_h, pool_w, padding=0, stride=self.stride
        )

        dX = dX.reshape(self.input_shape)

        return dX



# relative
from .network_msg_registry import NetworkMessageRegistry
from .node_service import NodeServiceClass


class NetworkServiceClass(NodeServiceClass):
    registry_type = NetworkMessageRegistry



# stdlib
from typing import Dict
from typing import List
from typing import Optional

# third party
from nacl.signing import VerifyKey

# relative
from ......grid import GridURL
from .....common.serde.serializable import serializable
from ....domain_msg_registry import DomainMessageRegistry
from ....network_msg_registry import NetworkMessageRegistry
from ....node_service import NodeServiceInterface
from ...permissions.user_permissions import NoRestriction
from ...permissions.user_permissions import UserIsOwner
from ..generic_payload.syft_message import NewSyftMessage as SyftMessage
from ..generic_payload.syft_message import ReplyPayload
from ..generic_payload.syft_message import RequestPayload
from .route_update import RouteUpdate

# AddRoute Messages
# Step 1: InitiateRouteUpdateToNodeMessage
# Step 2: NotifyNodeWithRouteUpdateMessage
# Step 3: VerifyRouteUpdateMessage

# for example a user would call step 1 from their python client to their domain, their
# domain would call step 2 on a network client, which would then in turn call step 3
# back on the domains new route and then return back through all the steps back to
# the user


@serializable(recursive_serde=True)
class InitiateRouteUpdateToNodeMessage(
    SyftMessage, DomainMessageRegistry, NetworkMessageRegistry
):
    permissions = [NoRestriction | UserIsOwner]  # UserIsOwner not working

    # Pydantic Inner class to define expected request payload fields.
    class Request(RequestPayload):
        """Payload fields and types used during a Request."""

        target_node_url: str
        route_update: dict

    # Pydantic Inner class to define expected reply payload fields.
    class Reply(ReplyPayload):
        """Payload fields and types used during a Response."""

        message: str = "Node route updated."

    request_payload_type = Request
    reply_payload_type = Reply

    def run(  # type: ignore
        self, node: NodeServiceInterface, verify_key: Optional[VerifyKey] = None
    ) -> ReplyPayload:  # type: ignore
        """Validates the request parameters and sends a ExchangeCredentialsWithNodeMessage.

        Args:
            node (NodeServiceInterface): Node either Domain or Network.
            verify_key (Optional[VerifyKey], optional): User signed verification key. Defaults to None.

        Raises:
            NodeTimeout: If the destination node is unavailable
            NodeError: If the destination node says there was an issue

        Returns:
            ReplyPayload: Message on successful exchange.
        """

        # TODO: get the client from the node and use a hashmap of uuid and / or url
        # to cache existing client objects to target nodes
        # get client for target node
        # allow uuid so user client doesn't always need to obtain a network client

        # relative
        from ......grid.client.client import connect

        # we may be trying to call another local test host like localhost so we need
        # to also call as_container_host
        target_url = (
            GridURL.from_url(self.payload.target_node_url)
            .with_path("/api/v1")
            .as_container_host(container_host=node.settings.CONTAINER_HOST)
        )

        # we use our local keys so that signing and verification matches our node
        target_client = connect(url=target_url, timeout=10, user_key=node.signing_key)

        # extract route update to send on to target node
        route_update = RouteUpdate(**self.payload.route_update)

        # see NotifyNodeWithRouteUpdateMessage below
        # TODO: Add retry mechanism with multiple routes in the above client
        _ = target_client.networking.notify_node_with_route_update(
            route_update=route_update
        )

        # check response coming back from target node before responding back to user

        return self.Reply()


@serializable(recursive_serde=True)
class NotifyNodeWithRouteUpdateMessage(
    SyftMessage, DomainMessageRegistry, NetworkMessageRegistry
):
    # TODO: Change to known verify_key of domain UserIsForeignNodeOwner?
    permissions = [NoRestriction]

    # Pydantic Inner class to define expected request payload fields.
    class Request(RequestPayload):
        """Payload fields and types used during a Request."""

        route_update: dict

    # Pydantic Inner class to define expected reply payload fields.
    class Reply(ReplyPayload):
        """Payload fields and types used during a Response."""

        message: str = "Node route has been updated."

    request_payload_type = Request
    reply_payload_type = Reply

    def run(  # type: ignore
        self, node: NodeServiceInterface, verify_key: Optional[VerifyKey] = None
    ) -> ReplyPayload:  # type: ignore
        """Validates the request parameters and exchanges credentials.

        Args:
            node (NodeServiceInterface): Node either Domain or Network.
            verify_key (Optional[VerifyKey], optional): User signed verification key. Defaults to None.

        Raises:
            NodeTimeout: If the destination node is unavailable
            NodeError: If the destination node says there was an issue

        Returns:
            ReplyPayload: Message on successful user creation.
        """

        # extract route update to send on to target node
        route_update = RouteUpdate(**self.payload.route_update)

        # check credentials for node asking to update routes has already been added
        # and matches the RouteUpdate request
        node_row = node.node.validate_id_and_key(
            node_uid=route_update.source_node_uid, verify_key=verify_key
        )

        if node_row is None:
            raise Exception("No node with that verify_key exists")

        # check that the url isn't already assigned to another node
        valid = node.node_route.validate_route_update(
            node_row=node_row, route_update=route_update, verify_key=verify_key
        )
        if not valid:
            raise Exception("host_or_ip and port are already assigned to another node")

        # check url works
        if route_update.autodetect is True:
            # autodetect url and try
            raise NotImplementedError()
        elif route_update.source_node_url:
            # create url
            source_node_url = GridURL.from_url(route_update.source_node_url)
        else:
            raise Exception("Invalid RouteUpdate", route_update)

        # TODO: get the client from the node and use a hashmap of uuid and / or url
        # to cache existing client objects to target nodes
        # get client for target node
        # allow uuid so user client doesn't always need to obtain a network client

        # relative
        from ......grid.client.client import connect

        # we may be trying to call another local test host like localhost so we need
        # to also call as_container_host
        target_url = source_node_url.with_path("/api/v1").as_container_host(
            container_host=node.settings.CONTAINER_HOST
        )

        # we use our local keys so that signing and verification matches our node
        target_client = connect(url=target_url, timeout=10, user_key=node.signing_key)

        # try to connect to the node and make sure its the one we expect
        signed_response = target_client.networking.verify_route()

        # since we're getting back the SignedMessage it can't hurt to check once more
        if not signed_response.is_valid:
            raise Exception(
                "Response was signed by a fake key or was corrupted in transit."
            )

        if signed_response.verify_key != verify_key:
            raise Exception("Request key doesn't match verify response key.")

        # save route to node_route table
        node.node_route.update_route(node_row=node_row, route_update=route_update)

        return self.Reply()


@serializable(recursive_serde=True)
class VerifyRouteUpdateMessage(
    SyftMessage, DomainMessageRegistry, NetworkMessageRegistry
):
    # TODO: Change to known verify_key of domain UserIsForeignNodeOwner?
    permissions = [NoRestriction]

    # Pydantic Inner class to define expected request payload fields.
    class Request(RequestPayload):
        """Payload fields and types used during a Request."""

        route_update: dict

    # Pydantic Inner class to define expected reply payload fields.
    class Reply(ReplyPayload):
        """Payload fields and types used during a Response."""

        message: str = "Node route has been updated."

    request_payload_type = Request
    reply_payload_type = Reply

    def run(  # type: ignore
        self, node: NodeServiceInterface, verify_key: Optional[VerifyKey] = None
    ) -> ReplyPayload:  # type: ignore
        """Validates the request parameters and exchanges credentials.

        Args:
            node (NodeServiceInterface): Node either Domain or Network.
            verify_key (Optional[VerifyKey], optional): User signed verification key. Defaults to None.

        Raises:
            NodeTimeout: If the destination node is unavailable
            NodeError: If the destination node says there was an issue

        Returns:
            ReplyPayload: Message on successful user creation.
        """

        print("Responding to VerifyRouteUpdateMessage")

        return self.Reply()


# RoutesListMessage
# Step 1: RoutesListMessage
# If the user has the right key they can manage their routes


@serializable(recursive_serde=True)
class RoutesListMessage(SyftMessage, DomainMessageRegistry, NetworkMessageRegistry):
    # TODO: Change to known verify_key of domain UserIsForeignNodeOwner?
    permissions = [NoRestriction]

    # Pydantic Inner class to define expected request payload fields.
    class Request(RequestPayload):
        """Payload fields and types used during a Request."""

        pass

    # Pydantic Inner class to define expected reply payload fields.
    class Reply(ReplyPayload):
        """Payload fields and types used during a Response."""

        routes_list: List[Dict]

    request_payload_type = Request
    reply_payload_type = Reply

    def run(  # type: ignore
        self, node: NodeServiceInterface, verify_key: Optional[VerifyKey] = None
    ) -> ReplyPayload:  # type: ignore
        """Validates the request parameters and exchanges credentials.

        Args:
            node (NodeServiceInterface): Node either Domain or Network.
            verify_key (Optional[VerifyKey], optional): User signed verification key. Defaults to None.

        Raises:
            NodeTimeout: If the destination node is unavailable
            NodeError: If the destination node says there was an issue

        Returns:
            ReplyPayload: Message on successful user creation.
        """

        node_row = node.node.get_node_for(verify_key=verify_key)
        if not node_row:
            raise Exception("There is no node for this verify_key")

        routes = node.node_route.get_routes(node_row=node_row)
        routes_list = [row._asdict() for row in routes]
        return self.Reply(routes_list=routes_list)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/tensor/share_tensor.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    recursive_serde_pb2 as proto_dot_core_dot_common_dot_recursive__serde__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b"\n$proto/core/tensor/share_tensor.proto\x12\x10syft.core.tensor\x1a'proto/core/common/recursive_serde.proto\"\x88\x01\n\x0bShareTensor\x12\r\n\x05\x63hild\x18\x01 \x01(\x0c\x12\x0c\n\x04rank\x18\x02 \x01(\r\x12\x11\n\tseed_przs\x18\x03 \x01(\r\x12\x11\n\tring_size\x18\x04 \x01(\x0c\x12\x36\n\x0cparties_info\x18\x05 \x03(\x0b\x32 .syft.core.common.RecursiveSerdeb\x06proto3"
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.core.tensor.share_tensor_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _SHARETENSOR._serialized_start = 100
    _SHARETENSOR._serialized_end = 236
# @@protoc_insertion_point(module_scope)



# stdlib
import json
from typing import Dict
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from typing_extensions import final

# relative
from ...... import serialize
from ......proto.grid.messages.network_search_messages_pb2 import (
    NetworkSearchMessage as NetworkSearchMessage_PB,
)
from ......proto.grid.messages.network_search_messages_pb2 import (
    NetworkSearchResponse as NetworkSearchResponse_PB,
)
from .....common.message import ImmediateSyftMessageWithReply
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.serde.deserialize import _deserialize
from .....common.serde.serializable import serializable
from .....common.uid import UID
from .....io.address import Address


@serializable()
@final
class NetworkSearchMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        content: Dict,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.content = content

    def _object2proto(self) -> NetworkSearchMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: NetworkSearchMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return NetworkSearchMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            content=json.dumps(self.content),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: NetworkSearchMessage_PB,
    ) -> "NetworkSearchMessage":
        """Creates a NetworkSearchMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: NetworkSearchMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return NetworkSearchMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=json.loads(proto.content),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return NetworkSearchMessage_PB


@serializable()
@final
class NetworkSearchResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        status_code: int,
        content: Dict,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.status_code = status_code
        self.content = content

    def _object2proto(self) -> NetworkSearchResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return NetworkSearchResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            status_code=self.status_code,
            content=json.dumps(self.content),
        )

    @staticmethod
    def _proto2object(
        proto: NetworkSearchResponse_PB,
    ) -> "NetworkSearchResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return NetworkSearchResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            status_code=proto.status_code,
            content=json.loads(proto.content),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return NetworkSearchResponse_PB



# future
from __future__ import annotations

# stdlib
from copy import deepcopy
import functools
from functools import lru_cache
import operator
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import Tuple
from typing import Union

# third party
import numpy as np
import torch

# syft absolute
import syft as sy

# relative
from . import utils
from .... import logger
from ....grid import GridURL
from ....lib.numpy.array import capnp_deserialize
from ....lib.numpy.array import capnp_serialize
from ....proto.core.tensor.share_tensor_pb2 import ShareTensor as ShareTensor_PB
from ...common.serde.capnp import CapnpModule
from ...common.serde.capnp import chunk_bytes
from ...common.serde.capnp import combine_bytes
from ...common.serde.capnp import get_capnp_schema
from ...common.serde.capnp import serde_magic_header
from ...common.serde.deserialize import _deserialize as deserialize
from ...common.serde.serializable import serializable
from ...common.serde.serialize import _serialize as serialize
from ...smpc.store.crypto_store import CryptoStore
from ..config import DEFAULT_RING_SIZE
from ..fixed_precision_tensor import FixedPrecisionTensor
from ..passthrough import PassthroughTensor  # type: ignore

if TYPE_CHECKING:
    # relative
    from ..tensor import Tensor


METHODS_FORWARD_ALL_SHARES = {
    "repeat",
    "copy",
    "diagonal",
    "flatten",
    "transpose",
    "partition",
    "resize",
    "ravel",
    "compress",
    "reshape",
    "squeeze",
    "swapaxes",
    "__pos__",
    "__neg__",
    "take",
    "choose",
    "cumsum",
    "trace",
}
INPLACE_OPS = {"resize", "put"}
RING_SIZE_TO_OP = {
    2: {
        "add": operator.xor,
        "sub": operator.xor,
        "mul": operator.and_,
        "lt": operator.lt,
        "gt": operator.gt,
        "ge": operator.ge,
        "le": operator.le,
        "eq": operator.eq,
        "ne": operator.ne,
    },
    2
    ** 32: {
        "add": operator.add,
        "sub": operator.sub,
        "mul": operator.mul,
        "matmul": operator.matmul,
        "lt": operator.lt,
        "gt": operator.gt,
        "ge": operator.ge,
        "le": operator.le,
        "eq": operator.eq,
        "ne": operator.ne,
    },
    2
    ** 64: {
        "add": operator.add,
        "sub": operator.sub,
        "mul": operator.mul,
        "matmul": operator.matmul,
        "lt": operator.lt,
        "gt": operator.gt,
        "ge": operator.ge,
        "le": operator.le,
        "eq": operator.eq,
        "ne": operator.ne,
    },
}

CACHE_CLIENTS: Dict[str, Any] = {}


def populate_store(*args: List[Any], **kwargs: Dict[Any, Any]) -> None:
    ShareTensor.crypto_store.populate_store(*args, **kwargs)  # type: ignore
    return None


@serializable(capnp_bytes=True)
class ShareTensor(PassthroughTensor):
    crypto_store = CryptoStore()

    __slots__ = (
        "rank",
        "ring_size",
        "clients",  # clients connections
        "min_value",
        "max_value",
        "generator_przs",
        # Only ShareTensors with seed_przs could be sent over the wire
        "seed_przs",
        "parties_info",
        "nr_parties",
    )

    def __init__(
        self,
        rank: int,
        parties_info: List[GridURL],
        ring_size: int,
        seed_przs: int = 42,
        clients: Optional[List[Any]] = None,
        value: Optional[Any] = None,
        init_clients: bool = False,
    ) -> None:
        # TODO: Ring size needs to be changed to 2^64 (or other specific sizes)
        self.rank = rank
        self.ring_size = ring_size
        self.nr_parties = len(parties_info)
        self.parties_info = parties_info
        self.clients = []
        if clients is not None:
            self.clients = clients
        elif init_clients:  # type: ignore
            self.clients = ShareTensor.login_clients(parties_info=parties_info)

        self.min_value, self.max_value = ShareTensor.compute_min_max_from_ring(
            self.ring_size
        )

        # This should be set only in the deserializer
        self.generator_przs = None
        self.seed_przs = seed_przs
        super().__init__(value)

    @staticmethod
    def login_clients(parties_info: List[GridURL]) -> Any:
        clients = []
        for party_info in parties_info:
            # if its localhost change it to a host that resolves outside the container
            external_host_info = party_info.as_container_host()
            client = CACHE_CLIENTS.get(str(external_host_info), None)

            if client is None:
                # default cache to true, here to prevent multiple logins
                # due to gevent monkey patching, context switch is done during
                # during socket connection initialization.
                CACHE_CLIENTS[str(external_host_info)] = True
                # TODO: refactor to use a guest account
                client = sy.login(  # nosec
                    url=external_host_info,
                    email="info@openmined.org",
                    password="changethis",
                    port=external_host_info.port,
                    verbose=False,
                )
                CACHE_CLIENTS[str(external_host_info)] = client
            clients.append(client)
        return clients

    def __getitem__(self, item: Union[str, int, slice]) -> ShareTensor:
        return ShareTensor(
            rank=self.rank,
            parties_info=self.parties_info,
            ring_size=self.ring_size,
            value=self.child[item],
            clients=self.clients,
        )

    def copy_tensor(self) -> ShareTensor:
        return ShareTensor(
            value=self.child,
            rank=self.rank,
            parties_info=self.parties_info,
            ring_size=self.ring_size,
            seed_przs=self.seed_przs,
            clients=self.clients,
        )

    @staticmethod
    @lru_cache(32)
    def compute_min_max_from_ring(
        ring_size: int = DEFAULT_RING_SIZE,
    ) -> Tuple[int, int]:
        if ring_size == 2:
            min_value, max_value = 0, 1
        else:
            min_value = (-ring_size) // 2
            max_value = (ring_size) // 2 - 1
        return min_value, max_value

    @staticmethod
    @lru_cache(maxsize=None)
    def get_op(ring_size: int, op_str: str) -> Callable[..., Any]:
        """Returns method attribute based on ring_size and op_str.
        Args:
            ring_size (int): Ring size
            op_str (str): Operation string.
        Returns:
            op (Callable[...,Any]): The operation method for the op_str.
        Raises:
            ValueError : If invalid ring size or op_str is given as input.
        """
        ops = RING_SIZE_TO_OP.get(ring_size, None)

        if ops is None:
            raise ValueError(f"Do not have operations for ring size {ring_size}")

        op = ops.get(op_str, None)
        if op is None:
            raise ValueError(
                f"Operator {op_str} does not exist for ring size {ring_size}"
            )

        return op

    """ TODO: Remove this -- we would use generate_przs since the scenario we are testing is that
    the secret is remotly
    @staticmethod
    def generate_shares(secret, nr_shares, ring_size=2 ** 64):
        from .fixed_precision_tensor import FixedPrecisionTensor

        if not isinstance(secret, (int, FixedPrecisionTensor)):
            secret = FixedPrecisionTensor(value=secret)

        shape = secret.shape
        min_value, max_value = ShareTensor.compute_min_max_from_ring(ring_size)

        generator_shares = np.random.default_rng()

        random_shares = []
        for i in range(nr_shares):
            random_value = generator_shares.integers(
                low=min_value, high=max_value, size=shape
            )
            fpt_value = FixedPrecisionTensor(value=random_value)
            random_shares.append(fpt_value)

        shares_fpt = []
        for i in range(nr_shares):
            if i == 0:
                share = value = random_shares[i]
            elif i < nr_shares - 1:
                share = random_shares[i] - random_shares[i - 1]
            else:
                share = secret - random_shares[i - 1]

            shares_fpt.append(share)

        # Add the ShareTensor class between them
        shares = []
        for rank, share_fpt in enumerate(shares_fpt):
            share_fpt.child = ShareTensor(rank=rank, value=share_fpt.child)
            shares.append(share_fpt)

        return shares
    """

    @staticmethod
    def generate_przs(
        value: Any,
        shape: Tuple[int, ...],
        rank: int,
        parties_info: List[GridURL],
        ring_size: Union[int, str] = DEFAULT_RING_SIZE,
        seed_przs: Optional[int] = None,
        generator_przs: Optional[Any] = None,
        init_clients: bool = True,
    ) -> Tensor:
        # relative
        from ..tensor import Tensor

        ring_size = int(ring_size)
        nr_parties = len(parties_info)

        # Try:
        # 1. First get numpy type if secret is numpy and obtain ring size from there
        # 2. If not get the type from the ring size

        numpy_type = None
        ring_size_final = None

        ring_size_from_type = utils.TYPE_TO_RING_SIZE.get(
            getattr(value, "dtype", None), None
        )
        if ring_size_from_type is None:
            logger.warning(f"Could not get ring size from {value}")
        else:
            ring_size_final = ring_size_from_type
            numpy_type = value.dtype

        if numpy_type is None:
            numpy_type = utils.RING_SIZE_TO_TYPE.get(ring_size, None)
            ring_size_final = ring_size

        if numpy_type is None:
            raise ValueError(f"Ring size {ring_size} not known how to be treated")

        if (seed_przs is None) == (generator_przs is None):
            raise ValueError("Only seed_przs or generator should be populated")

        if value is None:
            value = Tensor(np.zeros(shape, dtype=numpy_type))

        # TODO: Sending the seed and having each party generate the shares is not safe
        # Since the parties would know some of the other parties shares (this might not impose a risk
        # when shares are not sent between parties -- like private addition/subtraction, but it might
        # impose for multiplication
        # The secret holder should generate the shares and send them to the other parties
        if generator_przs:
            generator_shares = generator_przs
        else:
            generator_shares = np.random.default_rng(seed_przs)

        if isinstance(value.child, (ShareTensor, FixedPrecisionTensor)):
            value = value.child

        share = ShareTensor(
            value=value.child,
            rank=rank,
            parties_info=parties_info,
            seed_przs=seed_przs,  # type: ignore #TODO:Inspect as we could pass none.
            init_clients=init_clients,
            ring_size=ring_size_final,  # type: ignore
        )

        share.generator_przs = generator_shares
        shares = [
            generator_shares.integers(
                low=share.min_value,
                high=share.max_value,
                size=shape,
                endpoint=True,
                dtype=numpy_type,
            )
            for _ in range(nr_parties)
        ]

        op = ShareTensor.get_op(ring_size_final, "sub")
        przs_share = op(shares[rank], shares[(rank + 1) % nr_parties])
        share.child = op(share.child, przs_share)
        res = Tensor(share)

        return res

    @staticmethod
    def generate_przs_on_dp_tensor(
        value: Optional[Any],
        shape: Tuple[int],
        rank: int,
        parties_info: List[GridURL],
        seed_przs: int,
        share_wrapper: Any,
        ring_size: Union[int, str] = DEFAULT_RING_SIZE,
    ) -> PassthroughTensor:
        ring_size = int(ring_size)
        if hasattr(value, "child"):
            value.child.child = FixedPrecisionTensor(value.child.child)  # type: ignore

        if value is not None:
            share = ShareTensor.generate_przs(
                value=value.child,
                shape=shape,
                rank=rank,
                parties_info=parties_info,
                seed_przs=seed_przs,
                ring_size=ring_size,
            )
        else:
            share = ShareTensor.generate_przs(
                value=value,
                shape=shape,
                rank=rank,
                parties_info=parties_info,
                seed_przs=seed_przs,
                ring_size=ring_size,
            )
        # relative
        from ..autodp.gamma_tensor import GammaTensor
        from ..autodp.phi_tensor import PhiTensor

        if isinstance(share_wrapper.child, (PhiTensor, GammaTensor)):
            share_wrapper.child.child.child = share.child
        else:
            share_wrapper.child.child = share.child

        return share_wrapper

    @staticmethod
    def sanity_check(
        share: Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]
    ) -> None:
        """Check type for share

        Args:
            share (Union[int, float, ShareTensor, np.ndarray, torch.Tensor]): value to check

        Raises:
            ValueError: if type is not supported
        """
        if isinstance(share, float):
            raise ValueError("Type float not supported yet!")

        if isinstance(share, np.ndarray) and (
            not np.issubdtype(share.dtype, np.integer)
            and share.dtype != np.dtype("bool")
        ):
            raise ValueError(
                f"NPArray should have type int or bool, but found {share.dtype}"
            )

        if isinstance(share, torch.Tensor) and torch.is_floating_point(share):
            raise ValueError("Torch tensor should have type int, but found float")

    def apply_function(
        self,
        y: Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"],
        op_str: str,
    ) -> "ShareTensor":
        """Apply a given operation.

        Args:
            y (Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]): tensor to apply the operator.
            op_str (str): Operator.

        Returns:
            ShareTensor: Result of the operation.
        """
        ShareTensor.sanity_check(y)

        op = ShareTensor.get_op(self.ring_size, op_str)
        numpy_type = utils.RING_SIZE_TO_TYPE.get(self.ring_size, None)
        if numpy_type is None:
            raise ValueError(f"Do not know numpy type for ring size {self.ring_size}")

        if isinstance(y, ShareTensor):
            utils.get_ring_size(self.ring_size, y.ring_size)  # sanity check
            value = op(self.child, y.child)
        else:
            if op_str in {"add", "sub"}:
                # TODO: Converting y to numpy because doing "numpy op torch tensor" raises exception
                value = (
                    op(self.child, np.array(y, numpy_type))
                    if self.rank == 0
                    else deepcopy(self.child)
                )
            elif op_str in ["mul", "matmul", "lt"]:
                value = op(self.child, np.array(y, numpy_type))
            else:
                raise ValueError(f"{op_str} not supported")

        res = self.copy_tensor()
        res.child = value
        return res

    def add(
        self, y: Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]
    ) -> "ShareTensor":
        """Apply the "add" operation between "self" and "y".

        Args:
            y (Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]): self + y

        Returns:
            ShareTensor. Result of the operation.
        """
        new_share = self.apply_function(y, "add")
        return new_share

    def sub(
        self, y: Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]
    ) -> "ShareTensor":
        """Apply the "sub" operation between "self" and "y".

        Args:
            y (Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]): self - y

        Returns:
            ShareTensor. Result of the operation.
        """
        new_share = self.apply_function(y, "sub")
        return new_share

    def rsub(
        self, y: Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]
    ) -> "ShareTensor":
        """Apply the "rsub" operation between "self" and "y"

        Args:
            y (Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]): y - self

        Returns:
            ShareTensor. Result of the operation.
        """
        new_self = self.mul(-1)
        new_share = new_self.apply_function(y, "add")
        return new_share

    def mul(
        self, y: Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]
    ) -> "ShareTensor":
        """Apply the "mul" operation between "self" and "y".

        Args:
            y (Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]): self * y

        Returns:
            ShareTensor. Result of the operation.
        """
        # relative
        from ...node.common.action.smpc_action_functions import private_mul

        if isinstance(y, ShareTensor):
            new_share = private_mul(self, y, "mul")
        else:
            new_share = self.apply_function(y, "mul")

        return new_share

    def truediv(self, y: Union[int, float, np.ndarray, "ShareTensor"]) -> "ShareTensor":
        """Apply the "division" operation between "self" and "y".

        Args:
            y (Union[int, float, np.ndarray, "ShareTensor"]): self / y

        Returns:
            ShareTensor. Result of the operation.
        """
        # relative
        from ...node.common.action.smpc_action_functions import public_divide

        if not isinstance(y, (int, np.integer)):
            raise ValueError("Current Division only works for integers")
        else:
            if self.ring_size != 2:
                new_share = public_divide(self, y)
            else:
                new_share = self.copy_tensor()

        return new_share

    def bit_decomposition(self, ring_size: Union[int, str], bitwise: bool) -> None:
        """Apply the "decomposition" operation on self

        Args:
            ring_size (int): Ring size to decompose the shares
            bitwise (bool): Flag for bitwise decomposition.

        Returns:
            ShareTensor. Result of the operation.
        """
        # relative
        from ...node.common.action.smpc_action_functions import _decomposition

        ring_size = int(ring_size)

        _decomposition(self, ring_size, bitwise)

    def matmul(
        self, y: Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]
    ) -> "ShareTensor":
        """Apply the "matmul" operation between "self" and "y".

        Args:
            y (Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]): self @ y

        Returns:
            ShareTensor. Result of the operation.
        """
        # relative
        from ...node.common.action.smpc_action_functions import private_mul

        if isinstance(y, ShareTensor):
            new_share = private_mul(self, y, "matmul")
        else:
            new_share = self.apply_function(y, "matmul")

        return new_share

    def rmatmul(self, y: torch.Tensor) -> "ShareTensor":
        """Apply the "rmatmul" operation between "y" and "self".

        Args:
            y (torch.Tensor): y @ self

        Returns:
            ShareTensor. Result of the operation.
        """
        if isinstance(y, ShareTensor):
            raise ValueError("Private matmul not supported yet")

        new_share = ShareTensor.apply_function(self, y, "matmul")
        return new_share

    def lt(self, y: Union[ShareTensor, np.ndarray]) -> "ShareTensor":
        """Apply the "lt" operation between "y" and "self".

        Args:
            y (Union[ShareTensor,np.ndarray]): self < y

        Returns:
            ShareTensor. Result of the operation.
        """
        # raise ValueError(
        #     "It should not reach this point since we generate SMPCAction for this"
        # )
        new_share = ShareTensor.apply_function(self, y, "lt")
        return new_share

    def gt(self, y: Union[ShareTensor, np.ndarray]) -> "ShareTensor":
        """Apply the "gt" operation between "y" and "self".

        Args:
            y (Union[ShareTensor,np.ndarray]): self > y

        Returns:
            ShareTensor. Result of the operation.
        """
        # raise ValueError(
        #     "It should not reach this point since we generate SMPCAction for this"
        # )
        new_share = ShareTensor.apply_function(self, y, "gt")
        return new_share

    def ge(self, y: Union[ShareTensor, np.ndarray]) -> "ShareTensor":
        """Apply the "ge" operation between "y" and "self".

        Args:
            y (Union[ShareTensor,np.ndarray]): self >= y

        Returns:
            ShareTensor. Result of the operation.
        """
        # raise ValueError(
        #     "It should not reach this point since we generate SMPCAction for this"
        # )
        new_share = ShareTensor.apply_function(self, y, "ge")
        return new_share

    def le(self, y: Union[ShareTensor, np.ndarray]) -> "ShareTensor":
        """Apply the "le" operation between "y" and "self".

        Args:
            y (Union[ShareTensor,np.ndarray]): self <= y

        Returns:
            ShareTensor. Result of the operation.
        """
        # raise ValueError(
        #     "It should not reach this point since we generate SMPCAction for this"
        # )
        new_share = ShareTensor.apply_function(self, y, "le")
        return new_share

    def ne(self, y: Union[ShareTensor, np.ndarray]) -> "ShareTensor":
        """Apply the "ne" operation between "y" and "self".

        Args:
            y (Union[ShareTensor,np.ndarray]): self != y

        Returns:
            ShareTensor. Result of the operation.
        """
        # raise ValueError(
        #     "It should not reach this point since we generate SMPCAction for this"
        # )
        new_share = ShareTensor.apply_function(self, y, "ne")
        return new_share

    def eq(self, other: Any) -> bool:
        """Equal operator.
        Check if "self" is equal with another object given a set of
            attributes to compare.
        Args:
            other (Any): Value to compare.
        Returns:
            bool: True if equal False if not.
        """
        # TODO: Rasswanth: Fix later after the comparison operation
        # relative
        # from .... import Tensor

        # if (
        #     isinstance(self.child, Tensor)
        #     and isinstance(other.child, Tensor)
        #     and (self.child != other.child).child.any()  # type: ignore
        # ):
        #     return False

        # if (
        #     isinstance(self.child, np.ndarray)
        #     and isinstance(other.child, np.ndarray)
        #     and (self.child != other.child).any()
        # ):
        #     return False

        # if self.rank != other.rank:
        #     return False

        # if self.ring_size != other.ring_size:
        #     return False

        # if self.nr_parties != other.nr_parties:
        #     return False

        # return True

        return self.child == other.child

    # TRASK: commenting out because ShareTEnsor doesn't appear to have .session_uuid or .config
    # def div(
    #     self, y: Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]
    # ) -> "ShareTensor":
    #     """Apply the "div" operation between "self" and "y".
    #
    #     Args:
    #         y (Union[int, float, torch.Tensor, np.ndarray, "ShareTensor"]): Denominator.
    #
    #     Returns:
    #         ShareTensor: Result of the operation.
    #
    #     Raises:
    #         ValueError: If y is not an integer or LongTensor.
    #     """
    #     if not isinstance(y, (int, torch.LongTensor)):
    #         raise ValueError("Div works (for the moment) only with integers!")
    #
    #     res = ShareTensor(session_uuid=self.session_uuid, config=self.config)
    #     # res = self.apply_function(y, "floordiv")
    #     res.tensor = self.tensor // y
    #     return res

    def bit_extraction(self, pos: int = 0) -> ShareTensor:
        """Extracts the bit at the specified position.

        Args:
            pos (int): position to extract bit.

        Returns:
            ShareTensor : extracted bits at specific position.

        Raises:
            ValueError: If invalid position is provided.
        """
        ring_bits = utils.get_nr_bits(self.ring_size)
        if pos < 0 or pos > ring_bits - 1:
            raise ValueError(
                f"Invalid position for bit_extraction: {pos}, must be in range:[0,{ring_bits-1}]"
            )
        shape = self.shape
        numpy_type = utils.RING_SIZE_TO_TYPE[self.ring_size]
        # logical shift
        bit_mask = np.ones(shape, dtype=numpy_type) << pos
        value = self.child & bit_mask
        value = value.astype(np.bool_)
        share = self.copy_tensor()
        share.child = value
        return share

    def concatenate(
        self, other: ShareTensor, *args: List[Any], **kwargs: Dict[str, Any]
    ) -> ShareTensor:
        res = self.copy()
        res.child = np.concatenate((self.child, other.child), *args, **kwargs)
        return res

    @staticmethod
    def hook_method(__self: ShareTensor, method_name: str) -> Callable[..., Any]:
        """Hook a framework method.

        Args:
            method_name (str): method to hook

        Returns:
            A hooked method
        """

        def method_all_shares(
            _self: ShareTensor, *args: List[Any], **kwargs: Dict[Any, Any]
        ) -> Any:

            share = _self.child

            method = getattr(share, method_name)
            new_share = method(*args, **kwargs)

            res = _self.copy_tensor()
            res.child = np.array(new_share)

            return res

        return functools.partial(method_all_shares, __self)

    def __getattribute__(self, attr_name: str) -> Any:
        if attr_name in METHODS_FORWARD_ALL_SHARES or attr_name in INPLACE_OPS:
            return ShareTensor.hook_method(self, attr_name)

        return object.__getattribute__(self, attr_name)

    # TODO: Add capnp serialization to ShareTensor
    def _object2proto(self) -> ShareTensor_PB:
        # This works only for unsigned types.
        length_rs = self.ring_size.bit_length()
        rs_bytes = self.ring_size.to_bytes((length_rs + 7) // 8, byteorder="big")

        proto_init_kwargs = {
            "rank": self.rank,
            "parties_info": [serialize(party) for party in self.parties_info],
            "seed_przs": self.seed_przs,
            "ring_size": rs_bytes,
        }

        proto_init_kwargs["child"] = serialize(self.child, to_bytes=True)

        return ShareTensor_PB(**proto_init_kwargs)

    @staticmethod
    def _proto2object(proto: ShareTensor_PB) -> "ShareTensor":
        init_kwargs = {
            "rank": proto.rank,
            "parties_info": [deserialize(party) for party in proto.parties_info],
            "seed_przs": proto.seed_przs,
            "ring_size": int.from_bytes(proto.ring_size, "big"),
        }

        init_kwargs["value"] = deserialize(proto.child, from_bytes=True)

        # init_kwargs["init_clients"] = True
        res = ShareTensor(**init_kwargs)
        generator_przs = np.random.default_rng(proto.seed_przs)
        res.generator_przs = generator_przs
        return res

    def _object2bytes(self) -> bytes:
        schema = get_capnp_schema(schema_file="share_tensor.capnp")

        st_struct: CapnpModule = schema.ShareTensor  # type: ignore
        st_msg = st_struct.new_message()
        # this is how we dispatch correct deserialization of bytes
        st_msg.magicHeader = serde_magic_header(type(self))

        # child of Share tensor could either be Python Scalar or np.ndarray
        if isinstance(self.child, np.ndarray) or np.isscalar(self.child):
            chunk_bytes(
                capnp_serialize(np.array(self.child), to_bytes=True), "child", st_msg
            )
            st_msg.isNumpy = True
        else:
            chunk_bytes(serialize(self.child, to_bytes=True), "child", st_msg)  # type: ignore
            st_msg.isNumpy = False

        st_msg.rank = self.rank
        st_msg.partiesInfo = serialize(self.parties_info, to_bytes=True)
        st_msg.seedPrzs = self.seed_przs
        st_msg.ringSize = str(self.ring_size)

        return st_msg.to_bytes_packed()

    @staticmethod
    def _bytes2object(buf: bytes) -> ShareTensor:
        schema = get_capnp_schema(schema_file="share_tensor.capnp")
        st_struct: CapnpModule = schema.ShareTensor  # type: ignore
        # https://stackoverflow.com/questions/48458839/capnproto-maximum-filesize
        MAX_TRAVERSAL_LIMIT = 2**64 - 1

        st_msg = st_struct.from_bytes_packed(
            buf, traversal_limit_in_words=MAX_TRAVERSAL_LIMIT
        )

        if st_msg.isNumpy:
            child = capnp_deserialize(combine_bytes(st_msg.child), from_bytes=True)
        else:
            child = deserialize(combine_bytes(st_msg.child), from_bytes=True)

        return ShareTensor(
            value=child,
            rank=st_msg.rank,
            parties_info=deserialize(st_msg.partiesInfo, from_bytes=True),
            seed_przs=st_msg.seedPrzs,
            ring_size=int(st_msg.ringSize),
        )

    __add__ = add
    __radd__ = add
    __sub__ = sub
    __rsub__ = rsub
    __mul__ = mul
    __rmul__ = mul
    __matmul__ = matmul
    __rmatmul__ = rmatmul
    __truediv__ = truediv
    __lt__ = lt
    __gt__ = gt
    __ge__ = ge
    __le__ = le
    __eq__ = eq
    __ne__ = ne



# future
from __future__ import annotations

# stdlib
import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

# third party
import ascii_magic
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
from pydantic import BaseSettings

# relative
from ...lib.python import String
from ...logger import error
from ..common.message import SignedImmediateSyftMessageWithReply
from ..common.message import SignedMessage
from ..common.message import SyftMessage
from ..common.uid import UID
from ..io.location import Location
from ..io.location import SpecificLocation
from .common.node import Node
from .common.node_manager.association_request_manager import AssociationRequestManager
from .common.node_manager.node_manager import NodeManager
from .common.node_manager.node_route_manager import NodeRouteManager
from .common.node_manager.role_manager import RoleManager
from .common.node_manager.user_manager import UserManager
from .common.node_service.association_request.association_request_service import (
    AssociationRequestService,
)
from .common.node_service.association_request.association_request_service import (
    AssociationRequestWithoutReplyService,
)
from .common.node_service.network_search.network_search_service import (
    NetworkSearchService,
)
from .common.node_service.node_setup.node_setup_messages import (
    CreateInitialSetUpMessage,
)
from .common.node_service.node_setup.node_setup_service import NodeSetupService
from .common.node_service.peer_discovery.peer_discovery_service import (
    PeerDiscoveryService,
)
from .common.node_service.ping.ping_service import PingService
from .common.node_service.request_receiver.request_receiver_messages import (
    RequestMessage,
)
from .common.node_service.role_manager.role_manager_service import RoleManagerService
from .common.node_service.user_manager.user_manager_service import UserManagerService
from .common.node_service.vpn.vpn_service import VPNConnectService
from .common.node_service.vpn.vpn_service import VPNJoinSelfService
from .common.node_service.vpn.vpn_service import VPNJoinService
from .common.node_service.vpn.vpn_service import VPNRegisterService
from .common.node_service.vpn.vpn_service import VPNStatusService
from .domain import Domain
from .domain_client import DomainClient
from .network_client import NetworkClient
from .network_service import NetworkServiceClass


class Network(Node):

    network: SpecificLocation

    child_type = Domain
    client_type = NetworkClient
    child_type_client_type = DomainClient

    def __init__(
        self,
        name: Optional[str],
        network: SpecificLocation = SpecificLocation(),
        domain: Optional[Location] = None,
        device: Optional[Location] = None,
        vm: Optional[Location] = None,
        signing_key: Optional[SigningKey] = None,
        verify_key: Optional[VerifyKey] = None,
        root_key: Optional[VerifyKey] = None,
        db_engine: Any = None,
        settings: Optional[BaseSettings] = None,
    ):
        super().__init__(
            name=name,
            network=network,
            domain=domain,
            device=device,
            vm=vm,
            signing_key=signing_key,
            verify_key=verify_key,
            db_engine=db_engine,
            settings=settings,
        )

        # share settings with the FastAPI application level
        self.settings = settings

        # specific location with name
        self.network = SpecificLocation(name=self.name)
        self.root_key = root_key

        # Database Management Instances
        self.users = UserManager(db_engine)
        self.roles = RoleManager(db_engine)
        self.node = NodeManager(db_engine)
        self.node_route = NodeRouteManager(db_engine)
        self.association_requests = AssociationRequestManager(db_engine)

        # Grid Network Services
        self.immediate_services_with_reply.append(AssociationRequestService)
        self.immediate_services_with_reply.append(NodeSetupService)
        self.immediate_services_with_reply.append(RoleManagerService)
        self.immediate_services_with_reply.append(UserManagerService)
        self.immediate_services_with_reply.append(VPNConnectService)
        self.immediate_services_with_reply.append(VPNJoinService)
        self.immediate_services_with_reply.append(VPNRegisterService)
        self.immediate_services_with_reply.append(VPNStatusService)
        self.immediate_services_with_reply.append(VPNJoinSelfService)
        self.immediate_services_with_reply.append(PingService)
        self.immediate_services_with_reply.append(NetworkSearchService)
        self.immediate_services_with_reply.append(PeerDiscoveryService)

        # TODO: New Service registration process
        self.immediate_services_with_reply.append(NetworkServiceClass)

        self.immediate_services_without_reply.append(
            AssociationRequestWithoutReplyService
        )

        self.requests: List[RequestMessage] = list()
        # available_device_types = set()
        # TODO: add available compute types

        # default_device = None
        # TODO: add default compute type

        self._register_services()
        self.request_handlers: List[Dict[Union[str, String], Any]] = []
        self.handled_requests: Dict[Any, float] = {}

        self.post_init()

    def post_init(self) -> None:
        super().post_init()
        self.set_node_uid()

    def initial_setup(  # nosec
        self,
        signing_key: SigningKey,
        first_superuser_name: str = "Jane Doe",
        first_superuser_email: str = "info@openmined.org",
        first_superuser_password: str = "changethis",
        first_superuser_budget: float = 5.55,
        domain_name: str = "BigHospital",
    ) -> Network:
        Node.set_keys(node=self, signing_key=signing_key)

        # Build Syft Message
        msg: SignedImmediateSyftMessageWithReply = CreateInitialSetUpMessage(
            address=self.address,
            name=first_superuser_name,
            email=first_superuser_email,
            password=first_superuser_password,
            domain_name=domain_name,
            budget=first_superuser_budget,
            reply_to=self.address,
            signing_key=signing_key,
        ).sign(signing_key=self.signing_key)

        # Process syft message
        _ = self.recv_immediate_msg_with_reply(msg=msg).message

        return self

    def loud_print(self) -> None:
        try:
            install_path = os.path.abspath(
                os.path.join(os.path.realpath(__file__), "../../../img/")
            )
            ascii_magic.to_terminal(
                ascii_magic.from_image_file(
                    img_path=install_path + "/pygrid.png", columns=83
                )
            )

            print(
                r"""
                                                        |\ |  _ |_      _   _ |
                                                        | \| (- |_ \)/ (_) |  |(
    """
            )
        except Exception:
            print("NETOWRK NODE (print fail backup)")

    @property
    def icon(self) -> str:
        return "🔗"

    @property
    def id(self) -> UID:
        return self.network.id

    def message_is_for_me(self, msg: Union[SyftMessage, SignedMessage]) -> bool:
        # this needs to be defensive by checking network_id NOT network.id or it breaks
        try:
            return msg.address.network_id == self.id and msg.address.domain is None
        except Exception as e:
            error(f"Error checking if {msg.pprint} is for me on {self.pprint}. {e}")
            return False



# stdlib
from typing import Any

# relative
from ...abstract.node import AbstractNodeClient
from ...enums import ResponseObjectEnum
from ..node_service.role_manager.role_manager_messages import CreateRoleMessage
from ..node_service.role_manager.role_manager_messages import DeleteRoleMessage
from ..node_service.role_manager.role_manager_messages import GetRoleMessage
from ..node_service.role_manager.role_manager_messages import GetRolesMessage
from ..node_service.role_manager.role_manager_messages import UpdateRoleMessage
from .request_api import RequestAPI


class RoleRequestAPI(RequestAPI):
    def __init__(self, client: AbstractNodeClient):
        super().__init__(
            client=client,
            create_msg=CreateRoleMessage,
            get_msg=GetRoleMessage,
            get_all_msg=GetRolesMessage,
            update_msg=UpdateRoleMessage,
            delete_msg=DeleteRoleMessage,
            response_key=ResponseObjectEnum.ROLE,
        )

    def __getitem__(self, key: int) -> Any:
        return self.get(role_id=key)

    def __delitem__(self, key: int) -> None:
        self.delete(role_id=key)



"""This module contains `Property` attribute representing property objects which
implements getter and setter objects."""

# stdlib
from typing import Any
from typing import Callable as CallableT
from typing import Optional
from typing import Tuple
from typing import Union

# relative
from .. import ast
from ..logger import traceback_and_raise


class Property(ast.attribute.Attribute):
    """Creates property objects which implements getter and setter objects.

    Each valid action on AST triggers GetSetPropertyAction.
    """

    def __init__(
        self,
        path_and_name: str,
        parent: ast.attribute.Attribute,
        object_ref: Optional[Any] = None,
        return_type_name: Optional[str] = None,
        client: Optional[Any] = None,
    ) -> None:
        """Base constructor for Property Attribute.

        Args:
            client: The client for which all computation is being executed.
            path_and_name: The path for the current node, e.g. `syft.lib.python.List`.
            object_ref: The actual python object for which the computation is being made.
            return_type_name: The given action's return type name, with its full path, in string format.
            parent: The parent node in the AST.
        """
        super().__init__(
            path_and_name=path_and_name,
            parent=parent,
            object_ref=object_ref,
            return_type_name=return_type_name,
            client=client,
        )

        self.is_static = False

    def __call__(
        self,
        *args: Tuple[Any, ...],
        **kwargs: Any,
    ) -> Optional[Union[Any, CallableT]]:
        """`Property` attribute is not callable.

        Args:
            *args: List of arguments.
            **kwargs: Keyword arguments.

        Raises:
            ValueError: If the function is called.
        """
        traceback_and_raise(ValueError("Property should never be called."))
